# -*- coding: utf-8 -*-
"""
Created on Fri Sep 14 15:24:46 2018

@author: ivoseverins
"""
# Use the following lines on Mac
# from sys import platform
# if platform == "darwin":
#     from matplotlib import use
#     use('WXAgg')
import PySide2
import os  # Miscellaneous operating system interfaces - to be able to switch from Mac to Windows
from pathlib import Path  # For efficient path manipulation

import tqdm
import yaml
import numpy as np
import pandas as pd
# import wx

###################################################
## To enable interactive plotting with PySide2 in PyCharm 2022.3
import PySide2
import sys
sys.modules['PyQt5'] = sys.modules['PySide2']
from matplotlib import use
use('Qt5Agg')
###################################################

import matplotlib.pyplot as plt  # Provides a MATLAB-like plotting framework
import xarray as xr
from collections import UserDict
import re
import tifffile

from papylio.file import File
# from papylio.molecule import Molecules
from papylio.file_collection import FileCollection
from papylio.plotting import histogram
from papylio.movie.movie import Movie
# from papylio.plugin_manager import PluginManager
# from papylio.plugin_manager import PluginMetaClass
from papylio.plugin_manager import plugins

import re  # Regular expressions
import warnings
from nd2reader import ND2Reader


# import matplotlib.pyplot as plt #Provides a MATLAB-like plotting framework
# import itertools #Functions creating iterators for efficient looping
# np.seterr(divide='ignore', invalid='ignore')
# import pandas as pd
# from threshold_analysis_v2 import stepfinder
# import pickle

class Configuration(UserDict):
    # Ruamel yaml parser may be better for preserving comments
    # https://sourceforge.net/projects/ruamel-yaml/
    def __init__(self, filepath):
        self.reload_block = 0
        self.filepath = Path(filepath)
        self.previous_file_modification_time = 0

        # Load custom config file or otherwise load the default config file
        if self.filepath.is_file():
            self.load()
        else:
            filepath = Path(__file__).with_name('default_configuration.yml')
            with filepath.open('r') as yml_file:
                self._data = yaml.load(yml_file, Loader=yaml.SafeLoader)
            # self.load(Path(__file__).with_name('default_configuration.yml'))
            self.save()

    @property
    def data(self):
        self.load()
        return self._data

    @data.setter
    def data(self, data):
        self._data = data

    @property
    def file_modification_time(self):
        return self.filepath.stat().st_mtime

    def __enter__(self):
        self.load()
        self.reload_block += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reload_block -= 1

    @property
    def reload(self):
        return self.reload_block == 0

    def __getitem__(self, item):
        return self.data[item]

    def load(self, filepath=None):
        if self.reload:
            file_modification_time = self.file_modification_time
            if file_modification_time != self.previous_file_modification_time:
                if filepath is None:
                    filepath = self.filepath
                with filepath.open('r') as yml_file:
                    self._data = yaml.load(yml_file, Loader=yaml.SafeLoader)
                self.previous_file_modification_time = file_modification_time

    def save(self):
        with self.filepath.open('w') as yml_file:
            yaml.dump(self._data, yml_file, sort_keys=False)

# from PyQt5.QtWidgets import QFileDialog
def get_QApplication():
    from PySide2 import QtWidgets
    app = QtWidgets.QApplication.instance()
    if app is None:
        # if it does not exist then a QApplication is created
        app = QtWidgets.QApplication([])
    return app

def get_path():
    # if not 'app' in globals().keys():
    #     global app
    #     app = wx.App(None)
    # dlg = wx.DirDialog(None, message="Choose a folder", defaultPath="")
    # if dlg.ShowModal() == wx.ID_OK:
    #     path = dlg.GetPath()
    # else:
    #     path = None
    # dlg.Destroy()

    app = get_QApplication()
    from PySide2.QtWidgets import QFileDialog, QMainWindow
    path = QFileDialog.getExistingDirectory(QMainWindow(), 'Choose directory')
    return path

@plugins
class Experiment:
    """ Main experiment class

    Class containing all the files in an experiment.
    In fact it can contain any collection of files.

    .. warning:: Only works with one or two channels.

    Attributes
    ----------
    name : str
        Experiment name based on the name of the main folder
    main_path : str
        Absolute path to the main experiment folder
    files : list of :obj:`File`
        Files
    import_all : bool
        If true, then all files in the main folder are automatically imported. \n
        If false, then files are detected, but not imported.
    """
    # TODO: Add presets for specific microscopes
    def __init__(self, main_path=None, channels=['g', 'r'], import_all=True):
        """Init method for the Experiment class

        Loads config file if it locates one in the main directory, otherwise it exports the default config file to the main directory.
        Scans all directory in the main directory recursively and imports all found files (if import_all is set to `True`).

        Parameters
        ----------
        main_path : str
            Absolute path to the main experiment folder
        channels : list of str
            Channels used in the experiment
        import_all : bool
            If true, then all files in the main folder are automatically imported. \n
            If false, then files are detected, but not imported.
        """
        if main_path is None:
            main_path = get_path()
            if main_path is None:
                raise ValueError('No folder selected')

        self.name = os.path.basename(main_path)
        self.main_path = Path(main_path).absolute()
        self.files = FileCollection()
        self.import_all = import_all

        self._channels = np.atleast_1d(np.array(channels))
        self._number_of_channels = len(channels)
        self._pairs = [[c1, c2] for i1, c1 in enumerate(channels) for i2, c2 in enumerate(channels) if i2 > i1]

        # Load custom config file or otherwise load the default config file
        self.configuration = Configuration(self.main_path.joinpath('config.yml'))

        os.chdir(main_path)

        # file_paths = self.find_file_paths()
        # self.add_files(file_paths, test_duplicates=False)

        with self.configuration:
            self.add_files(self.main_path, test_duplicates=False)

        self.common_image_corrections = xr.Dataset()
        self.load_darkfield_correction()
        self.load_flatfield_correction()

        # Find mapping file
        for file in self.files:
            if file.mapping is not None:
                file.use_mapping_for_all_files()
                break

        print('\nInitialize experiment: \n' + str(self.main_path))

    def __getstate__(self):
        d = self.__dict__.copy()
        # d.pop('files')
        d['files'] = []
        excluded_keys = ['files', 'sequencing_data', '_tile_mappings'] #TODO: Move sequencing related terms to sequencing.py
        d = {key: value for key, value in self.__dict__.items() if key not in excluded_keys}
        d['_do_not_update'] = None # This is for parallelization in Collection
        return d

    def __setstate__(self, dict):
        self.__dict__.update(dict)

    def __repr__(self):
        return (f'{self.__class__.__name__}({self.name})')

    @property
    def channels(self):
        """list of str : Channels used in the experiment.

        Setting the channels will automatically update pairs.
        """
        return self._channels

    @channels.setter
    def channels(self, channels):
        self._channels = np.atleast_1d(np.array(channels))
        self._number_of_channels = len(channels)
        self._pairs = [[c1, c2] for i1, c1 in enumerate(channels) for i2, c2 in enumerate(channels) if i2 > i1]

    @property
    def number_of_channels(self):
        """int : Number of channels used in the experiment (read-only)"""
        return self._number_of_channels

    @property
    def pairs(self):
        """list of list of str : List of channel pairs"""
        return self._pairs

    # @property
    # def molecules(self):
    #     """list of Molecule : List of all molecules in the experiment"""
    #     return Molecules.sum([file.molecules for file in self.files])

    @property
    def selectedFiles(self):
        """list of File : List of selected files"""
        return self.files[self.files.isSelected]

    # @property
    # def selectedMoleculesInSelectedFiles(self):
    #     """list of Molecule : List of selected molecules in selected files"""
    #     return [molecule for file in self.selectedFiles for molecule in file.selectedMolecules]

    # @property
    # def selectedMoleculesInAllFiles(self):
    #     """list of Molecule : List of selected molecules in all files"""
    #     return [molecule for file in self.files for molecule in file.selectedMolecules]

    @property
    def mapping_file(self):
        for file in self.files:
            if file.is_mapping_file:
                return file

    @property
    def analysis_path(self):
        analysis_path = self.main_path.joinpath('Analysis')
        analysis_path.mkdir(parents=True, exist_ok=True)
        return analysis_path

    @property
    def file_paths(self):
        return [file.relativeFilePath for file in self.files]

    @property
    def nc_file_paths(self):
        return [file.relativeFilePath.with_suffix('.nc') for file in self.files if '.nc' in file.extensions]

    def find_file_paths_and_extensions(self, paths):
        """Find unique files in all subfolders and add them to the experiment

        Get all files in all subfolders of the main_path and remove their suffix (extensions), and add them to the experiment.

        Note
        ----
        Non-relevant files are excluded e.g. files with underscores or 'Analysis' in their name, or files with dat, db,
        ini, py and yml extensions.

        Note
        ----
        Since sifx files made using spooling are all called 'Spooled files' the parent folder is used as file instead of the sifx file

        """

        if isinstance(paths, str) or isinstance(paths, Path):
            #paths = paths.glob('**/*')
                #'**/?*.*')  # At least one character in front of the extension to prevent using hidden folders

            # The following approach is faster than checking each file separately using is_file() for network drives. (not tested for regular drives)
            files_and_folders = set(paths.glob('**/*'))
            folders = set(paths.glob('**'))
            paths = files_and_folders - folders

        file_paths_and_extensions = \
            [[p.relative_to(self.main_path).with_suffix(''), p.suffix]
             for p in paths
             if (
                     # Use only files
                     #p.is_file() &
                     # Exclude stings in filename
                     all(name not in p.with_suffix('').name for name in
                         self.configuration['files']['excluded_names']) &
                     # Exclude strings in path
                     all(path not in str(p.relative_to(self.main_path).parent) for path in
                         self.configuration['files']['excluded_paths']) &
                     # Exclude hidden folders
                     ('.' not in [s[0] for s in p.parts]) &
                     # Exclude file extensions
                     (p.suffix[1:] not in self.configuration['files']['excluded_extensions'])
             )
             ]

        # TODO: Test spooled file and nd2 file import
        new_file_paths_and_extensions = []
        for i, (file_path, extensions) in enumerate(file_paths_and_extensions):
            if (file_path.name == 'Spooled files'):
                new_file_paths_and_extensions.append([file_path.parent, extensions])
                # file_paths_and_extensions[i, 0] = file_paths_and_extensions[i, 0].parent
            elif '.nd2' in extensions and not 'fov' in str(file_path):
                from papylio.movie.movie import Movie
                nd2_movie = Movie(file_path.with_suffix('.nd2'))
                if nd2_movie.number_of_fov > 1:  # if the file is nd2 with multiple field of views
                    for fov_id in range(nd2_movie.number_of_fov):
                        new_path = Path(str(file_path) + f'_fov{fov_id:03d}')
                        new_file_paths_and_extensions.append([new_path, extensions])
                        # fov_info['fov_chosen'] = fov_id
                        # new_file = File(new_path, self, fov_info=fov_info.copy())
                        # if new_file.extensions:
                        #     self.files.append(new_file)
                else:
                    new_file_paths_and_extensions.append([file_path, extensions])
            else:
                new_file_paths_and_extensions.append([file_path, extensions])
        file_paths_and_extensions = new_file_paths_and_extensions

        file_paths_and_extensions = np.array(file_paths_and_extensions)

        file_paths_and_extensions = file_paths_and_extensions[file_paths_and_extensions[:, 0].argsort()]
        unique_file_paths, indices = np.unique(file_paths_and_extensions[:, 0], return_index=True)
        extensions_per_filepath = np.split(file_paths_and_extensions[:, 1], indices[1:])

        return unique_file_paths, extensions_per_filepath

    def add_files(self, paths, test_duplicates=True):
        """Find unique files in all subfolders and add them to the experiment

        Get all files in all subfolders of the main_path and remove their suffix (extensions), and add them to the experiment.

        Note
        ----
        Non-relevant files are excluded e.g. files with underscores or 'Analysis' in their name, or files with dat, db,
        ini, py and yml extensions.

        Note
        ----
        Since sifx files made using spooling are all called 'Spooled files' the parent folder is used as file instead of the sifx file

        """
        file_paths_and_extensions = self.find_file_paths_and_extensions(paths)

        for file_path, extensions in tqdm.tqdm(zip(*file_paths_and_extensions), 'Import files',
                                               total=(len(file_paths_and_extensions[0]))):
            if not test_duplicates or (file_path.absolute().relative_to(self.main_path) not in self.file_paths):
                self.files.append(File(file_path, extensions, self))
            else:
                i = self.file_paths.find(file_path.absolute().relative_to(self.main_path))
                self.files[i].add_extensions(extensions)

        # nd2_file = list(self.main_path.glob(str(relativeFilePath) + '.nd2'))
        # fov_info = {'number_of_fov': 1}  # fov=Field of View
        # if nd2_file:  # check if the nd2 file has multiple fov data,
        #     fov_info = self.get_fov_from_nd2(nd2_file[0])
        #
        # if fov_info['number_of_fov'] > 1:  # if the file is nd2 with multiple field of views
        #     for fov_id in range(fov_info['number_of_fov']):
        #         new_path = Path(str(relativeFilePath) + f'_fov{fov_id:03d}')
        #         fov_info['fov_chosen'] = fov_id
        #         new_file = File(new_path, self, fov_info=fov_info.copy())
        #         if new_file.extensions:
        #             self.files.append(new_file)



    # def add_files(self, file_paths, test_duplicates=True):
    #     for file_path in file_paths:
    #         self.add_file(file_path, test_duplicates)
    #
    #     for file in self.files:
    #         if file.mapping is not None:
    #             file.use_mapping_for_all_files()
    #             break

    # def add_file(self, file_path, test_duplicates=True):
    #     """Add a file to the experiment
    #
    #     Add the file to the experiment only if the file object has found and imported relevant extensions .
    #     If the file is already present in experiment, then try to find and import new extensions.
    #
    #     Parameters
    #     ----------
    #     relativeFilePath : pathlib.Path or str
    #         Path with respect to the main experiment path
    #
    #     """
    #     # Perhaps move this conversion to relative file path to File
    #     relative_file_path = Path(file_path).absolute().relative_to(self.main_path)
    #
    #     # if there is no extension, add all files with the same name with all extensions
    #     # if there is an extension just add that file if the filename is the same
    #
    #     # Test whether file is already in experiment
    #     # for file in self.files:
    #     #     if file.relativeFilePath == relativeFilePath:
    #     #         file.findAndAddExtensions()
    #     #         break
    #     # else:
    #     #     new_file = File(relativeFilePath, self)
    #     #     if new_file.extensions:
    #     #         self.files.append(new_file)
    #
    #
    #     if not test_duplicates or (relative_file_path not in self.file_paths):
    #         self.files.append(File(relative_file_path, self))
    #     else:
    #         i = self.file_paths.find(file_path.absolute().relative_to(self.main_path))
    #         self.files[i].findAndAddExtensions()

    # def load_flatfield_correction(self):
    #     file_paths = list(self.main_path.glob('flatfield*'))
    #     if file_paths:
    #         movie = self.files[0].movie
    #         flatfield_correction = xr.DataArray(np.ones((movie.number_of_illuminations,
    #                                             movie.height, movie.width)), # perhaps make the movie width and height equal to the channel width and height
    #                               dims=('illumination', 'y', 'x'),
    #                               coords={'illumination': movie.illumination_indices})
    #         for file_path in file_paths:
    #             flatfield = tifffile.imread(file_path)
    #             _, _, illumination_indices, _ = movie.image_type_from_filename(file_path.name)
    #             flatfield_correction[dict(illumination=illumination_indices)] = flatfield
    #
    #         self.files.movie.flatfield_correction = flatfield_correction
    #     else:
    #         self.files.movie.flatfield_correction = None

    def determine_flatfield_and_darkfield_corrections(self, files, method='BaSiC', illumination_index=0, frame_index=0,
                                                      estimate_darkfield=True, **kwargs):
        from papylio.movie.basic_shading_correction import spatial_shading_correction

        darkfield, flatfield = spatial_shading_correction(files.movie, method=method,
                                                          illumination_index=illumination_index,
                                                          frame_index=frame_index,
                                                          estimate_darkfield=estimate_darkfield, **kwargs)
        if estimate_darkfield:
            tifffile.imwrite(self.main_path / f'darkfield_i{illumination_index}.tif', darkfield.astype('float32'), imagej=True)
            self.load_darkfield_correction()
        tifffile.imwrite(self.main_path / f'flatfield_i{illumination_index}.tif', flatfield.astype('float32'), imagej=True)
        self.load_flatfield_correction()

    def load_flatfield_correction(self):
        file_paths = list(self.main_path.glob('flatfield*'))
        if file_paths:
            movie = self.files[0].movie
            flatfield_correction = xr.DataArray(np.ones((movie.number_of_illuminations, movie.number_of_channels,
                                                         movie.channels[0].height, movie.channels[0].width)),
                                                # perhaps make the movie width and height equal to the channel width and height
                                                dims=('illumination', 'channel', 'y', 'x'),
                                                coords={'illumination': movie.illumination_indices,
                                                        'channel': movie.channel_indices})
            for file_path in file_paths:
                flatfield = tifffile.imread(file_path)
                image_info = Movie.image_info_from_filename(file_path.name)
                illumination_index = image_info['illumination_index']
                channel_indices = movie.channel_indices
                flatfield_correction[dict(illumination=illumination_index, channel=channel_indices)] = \
                    movie.separate_channels(flatfield)

            self.common_image_corrections['flatfield_correction'] = flatfield_correction
        else:
            self.common_image_corrections = self.common_image_corrections.drop_vars('flatfield_correction', errors='ignore')

        self.add_common_image_corrections_to_movies()

    def load_darkfield_correction(self):
        file_paths = list(self.main_path.glob('darkfield*'))
        if file_paths:
            movie = self.files[0].movie
            darkfield_correction = xr.DataArray(np.ones((movie.number_of_illuminations, movie.number_of_channels,
                                                         movie.channels[0].height, movie.channels[0].width)),
                                                # perhaps make the movie width and height equal to the channel width and height
                                                dims=('illumination', 'channel', 'y', 'x'),
                                                coords={'illumination': movie.illumination_indices,
                                                        'channel': movie.channel_indices})
            # for file_path in file_paths:
            darkfield = tifffile.imread(file_paths[0])
            for illumination_index in darkfield_correction.illumination.values:
                # image_info = Movie.image_info_from_filename(file_path.name)
                # illumination_index = image_info['illumination_index']
                channel_indices = movie.channel_indices
                darkfield_correction[dict(illumination=illumination_index, channel=channel_indices)] = \
                    movie.separate_channels(darkfield)

            self.common_image_corrections['darkfield_correction'] = darkfield_correction
        else:
            self.common_image_corrections = self.common_image_corrections.drop_vars('darkfield_correction', errors='ignore')

        self.add_common_image_corrections_to_movies()

    def add_common_image_corrections_to_movies(self):
        self.files.movie._common_corrections = self.common_image_corrections

    # def show_flatfield_and_darkfield_corrections(self, name='', save=True):
    #     pass

    # def load_darkfield_correction(self):
    #     file_paths = list(self.main_path.glob('darkfield*'))
    #     if file_paths:
    #         movie = self.files[0].movie
    #         darkfield_correction = xr.DataArray(np.zeros((movie.number_of_illuminations,
    #                                             movie.height, movie.width)), # perhaps make the movie width and height equal to the channel width and height
    #                                    dims=('illumination', 'y', 'x'),
    #                                    coords={'illumination': movie.illumination_indices})
    #         for file_path in file_paths:
    #             darkfield = tifffile.imread(file_path)
    #             _, _, illumination_indices, _ = movie.image_type_from_filename(file_path.name)
    #             darkfield_correction[dict(illumination=illumination_indices)] = darkfield
    #
    #         self.files.movie.darkfield_correction = darkfield_correction
    #     else:
    #         self.files.movie.darkfield_correction = None

    def histogram(self, axis=None, bins=100, parameter='E', molecule_averaging=False,
                  fileSelection=False, moleculeSelection=False, makeFit=False, export=False, **kwargs):
        """FRET histogram of all molecules in the experiment or a specified selection

        Parameters
        ----------
        axis : matplotlib.axis
            Axis to use for histogram plotting
        bins : int
            Number of bins
        parameter : str
            Parameter to be used for histogram I or E
        molecule_averaging : bool
            If True an time average of the trace is used
        fileSelection : bool
            If True the histogram is made only using selected files.
        moleculeSelection : bool
            If True the histogram is made only using selected molecules.
        makeFit : bool
            If True perform Gaussian fitting.
        export : bool
            If True the graph is exported.
        **kwargs
            Arbitrary keyword arguments.

        """
        # files = [file for file in exp.files if file.isSelected]
        # files = self.files

        if (fileSelection & moleculeSelection):
            molecules = [molecule for file in self.selectedFiles for molecule in file.selectedMolecules]
        elif (fileSelection & (not moleculeSelection)):
            molecules = [molecule for file in self.selectedFiles for molecule in file.molecules]
        elif ((not fileSelection) & moleculeSelection):
            molecules = [molecule for file in self.files for molecule in file.selectedMolecules]
        else:
            molecules = [molecule for file in self.files for molecule in file.molecules]

        histogram(molecules, axis=axis, bins=bins, parameter=parameter, molecule_averaging=molecule_averaging,
                  makeFit=makeFit, collection_name=self, **kwargs)
        if export: plt.savefig(self.main_path.joinpath(f'{self.name}_{parameter}_histogram').with_suffix('.png'))

    def boxplot_number_of_molecules(self):
        """Boxplot of the number of molecules in each file"""
        fig, ax = plt.subplots(figsize=(8, 1.5))
        pointCount = [len(file.molecules) for file in self.files]
        plt.boxplot(pointCount, vert=False, labels=[''], widths=(0.8))
        plt.xlabel('Count')
        plt.title('Molecules per file')
        plt.tight_layout()

        fig.savefig(self.main_path.joinpath('number_of_molecules.pdf'), bbox_inches='tight')
        fig.savefig(self.main_path.joinpath('number_of_molecules.png'), bbox_inches='tight')

    def select(self):
        """Simple method to look through all molecules in the experiment

        Plots a molecule. If enter is pressed the next molecule is shown.

        """
        for molecule in self.molecules:
            molecule.plot()
            input("Press enter to continue")

    def print_files(self):
        self.files.print()

    def plot_trace(self, files=None, query={}, **kwargs):
        # from papylio.trace_plot import TraceAnalysisFrame

        if files is None:
            files = self.files

        file_paths = [file.relativeFilePath.with_suffix('.nc') for file in files if '.nc' in file.extensions]

        with xr.open_mfdataset(file_paths, concat_dim='molecule', combine='nested') as ds:
            ds_sel = ds.query(query)  # HJ1_WT, HJ7_G116T
            # if not 'app' in globals().keys():
            #     global app
            #     app = wx.App(None)
            # app = wit.InspectableApp()
            # frame = TraceAnalysisFrame(None, ds_sel, "Sample editor")
            # frame.molecules = exp.files[1].molecules
            # print('test')
            # import wx.lib.inspection
            # wx.lib.inspection.InspectionTool().Show()
            # app.MainLoop()
            from papylio.trace_plot import TracePlotWindow
            TracePlotWindow(dataset=ds_sel, save_path=None, **kwargs)

    def export_number_of_molecules_per_file(self):
        df = pd.DataFrame(columns=['Number of molecules'])
        for i, file in enumerate(self.files):
            n = str(file.relativeFilePath)
            try:
                nms = file.number_of_molecules
            except FileNotFoundError:
                nms = -1
            df.loc[n] = nms
        df.to_excel(self.main_path.joinpath('number_of_molecules'))



import tqdm
import numpy as np
import netCDF4
from joblib import Parallel, delayed
from pathlib2 import Path


def get_dimension_size(filepath, dimension, with_selected_only=False, with_sequence_only=False):
    with netCDF4.Dataset(filepath) as ds:
        if with_selected_only:
            return ds['selected'][:].sum()
        elif with_sequence_only:
            # return (~(ds['sequence_aligned'][:] == b'-').all(axis=1)).sum()
            return (ds['sequence_tile'][:] > 0).sum()
        else:
            return ds.dimensions[dimension].size


def get_dimension_sizes(filepaths, dimension, with_selected_only=False, with_sequence_only=False):
    return Parallel(prefer="threads")(delayed(get_dimension_size)(filepath, dimension, with_selected_only, with_sequence_only)
                                      for filepath in filepaths)
    # return [get_dimension_size(filepath, 'molecule', with_sequence_only) for filepath in filepaths]


def merge_datasets(files_in, file_out, concat_dim, init_file=None, with_selected_only=False, with_sequence_only=False):
    # TODO: remove sequencing part, or move to the sequencing plugin
    if init_file is None:
        init_file = files_in[0]

    concat_dim_size = np.sum(get_dimension_sizes(files_in, concat_dim, with_selected_only, with_sequence_only))

    with netCDF4.Dataset(file_out, mode='w') as ds_out:
        with netCDF4.Dataset(init_file) as ds_in:
            init_dataset_like(ds_in, ds_out, concat_dim, concat_dim_size=concat_dim_size)

        start_index_out = 0
        for file_in in tqdm.tqdm(files_in):
            with netCDF4.Dataset(file_in) as ds_in:
                if with_selected_only:
                    _, start_index_out = append_to_dataset(ds_in, ds_out, concat_dim, selection_in=ds_in['selected'][:].astype(bool),
                                                           start_index_out=start_index_out)
                elif with_sequence_only:
                    # has_sequence = (~(ds_in['sequence_aligned'][:] == b'-').all(axis=1))
                    has_sequence = ds_in['sequence_tile'][:] > 0
                    _, start_index_out = append_to_dataset(ds_in, ds_out, concat_dim, selection_in=has_sequence, start_index_out=start_index_out)
                else:
                    _, start_index_out = append_to_dataset(ds_in, ds_out, concat_dim, selection_in=None, start_index_out=start_index_out)


def reorder_datasets_using_sequence_subset(files_in, folder_out, concat_dim):
    folder_out = Path(folder_out)
    folder_out.mkdir(exist_ok=False)
    for file_in in tqdm.tqdm(files_in):
        with netCDF4.Dataset(file_in) as ds_in:
            # selection = np.squeeze(ds['sequence_subset'][:].view('S8') != b'--------')
            sequence_subsets = ds_in['sequence_subset'][:]
            sequence_subsets.set_fill_value(b'-')
            sequence_subsets = sequence_subsets.filled()
            selection = (sequence_subsets != b'-').all(axis=1)
            indices = np.where(selection)[0]

            sequence_subsets = sequence_subsets.view(f'S{sequence_subsets.shape[1]}').astype('U').squeeze()

            for i in indices:
                sequence_subset = sequence_subsets[i]
                with netCDF4.Dataset(folder_out / (sequence_subset + '.nc'), mode='a') as ds_out:
                    if concat_dim not in ds_out.dimensions.keys():
                        init_dataset_like(ds_in, ds_out, concat_dim)
                    append_index_to_dataset(ds_in, ds_out, concat_dim, i)


def init_dataset_like(ds_in, ds_out, concat_dim, concat_dim_size=None):
    ds_out.createDimension(concat_dim, concat_dim_size)
    for name, dimension in ds_in.dimensions.items():
        if name != concat_dim:
            ds_out.createDimension(dimension.name, dimension.size)

    for name, variable in ds_in.variables.items():
        if '_FillValue' in variable.ncattrs():
            fill_value = variable.getncattr('_FillValue')
        else:
            fill_value = None

        ds_out.createVariable(variable.name, variable.datatype, dimensions=variable.dimensions,
                               fill_value=fill_value)

        for attr in variable.ncattrs():
            if attr != '_FillValue':
                ds_out[name].setncattr(attr, variable.getncattr(attr))

        if concat_dim not in variable.dimensions:
            ds_out[name][:] = ds_in[name][:]


def append_to_dataset(ds_in, ds_out, concat_dim, selection_in=None, start_index_out=None):
    if selection_in is None:
        selection_in = np.ones(ds_in.dimensions[concat_dim].size).astype(bool)
    if start_index_out is None:
        if ds_out.dimensions[concat_dim].isunlimited():
            start_index_out = ds_out.dimensions[concat_dim].size
        else:
            start_index_out = 0
    # end = start + ds.dimensions[concat_dim].size
    end_index_out = start_index_out + selection_in.sum()

    for name, variable in ds_in.variables.items():
        if concat_dim in variable.dimensions:
            if name == 'file':
                min_len = min(ds_in[name].shape[-1], ds_out[name].shape[-1])
                ds_out[name][start_index_out:end_index_out, :min_len] = ds_in[name][selection_in, :min_len]
            else:
                ds_out[name][start_index_out:end_index_out] = ds_in[name][selection_in]

    return start_index_out, end_index_out


def append_index_to_dataset(ds_in, ds_out, concat_dim, index_in):
    index_to = ds_out.dimensions[concat_dim].size
    for name, variable in ds_in.variables.items():
        if concat_dim in variable.dimensions:
            if name == 'file':
                min_len = min(ds_in[name].shape[-1], ds_out[name].shape[-1])
                ds_out[name][index_to, :min_len] = ds_in[name][index_in, :min_len]
            else:
                ds_out[name][index_to] = ds_in[name][index_in]



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 14 15:44:52 2018

@author: ivoseverins
"""
# import wx
# import wx.lib.mixins.inspection as wit
# import sys
# print('PyQt5', sys.modules.get("PyQt5.QtCore"))
# print('PySide2', sys.modules.get("PySide2.QtCore"))

###################################################
## To enable interactive plotting with PySide2 in PyCharm 2022.3
import PySide2
import sys
sys.modules['PyQt5'] = sys.modules['PySide2']
import matplotlib
matplotlib.use('Qt5Agg')
###################################################

import matplotlib.pyplot as plt

# from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
# from matplotlib.backends.backend_wxagg import NavigationToolbar2WxAgg as NavigationToolbar
# from matplotlib.backends.backend_qtagg import FigureCanvas
# from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

import numpy as np
from pathlib2 import Path

from PySide2.QtWidgets import QMainWindow, QPushButton, QWidget, QVBoxLayout, QHBoxLayout, QLineEdit, QCheckBox, QLabel
from PySide2.QtGui import QKeySequence
from PySide2.QtCore import Qt

import sys
import time

import numpy as np

# from matplotlib.backends.qt_compat import QtWidgets
from PySide2 import QtWidgets
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg, NavigationToolbar2QT as NavigationToolbar)
from matplotlib.figure import Figure


class TracePlotWindow(QWidget):
    def __init__(self, dataset=None, plot_variables=['intensity', 'FRET'],
                 ylims=[(0, 35000), (0, 1)], colours=[('g', 'r'), ('b')], width=14, height=None, save_path=None, parent=None,
                 show=True):

        if height is None:
            height = max(len(plot_variables) * 3.5, 9)

        from papylio.experiment import get_QApplication
        #TODO: Use selection only if it is present.
        app = get_QApplication()

        super().__init__()

        self.parent = parent

        self.setWindowTitle("Traces")

        self.plot_variables = plot_variables
        self.ylims = ylims
        self.colours = colours

        if save_path is None:
            self.save_path = save_path
        else:
            self.save_path = Path(save_path)

        self._dataset = dataset

        self.canvas = TracePlotCanvas(self, width=width, height=height, dpi=100)

        # Create toolbar, passing canvas as first parament, parent (self, the MainWindow) as second.
        toolbar = NavigationToolbar(self.canvas, self)

        layout = QVBoxLayout()

        layout_bar = QHBoxLayout()
        layout_bar.addWidget(toolbar, 0.5)

        self.molecule_index_field = QLineEdit()
        self.molecule_index_field.setFixedWidth(70)

        layout_bar.addWidget(self.molecule_index_field, 0.05)
        layout_bar.addWidget(QLabel(' out of '), 0.05)
        self.number_of_molecules_label = QLabel('0')
        self.number_of_molecules_label.setFixedWidth(70)
        layout_bar.addWidget(self.number_of_molecules_label, 0.15)
        self._selection_state = 1
        self.selected_molecules_checkbox = QCheckBox()
        self.selected_molecules_checkbox.setTristate(True)
        self.selected_molecules_checkbox.setCheckState(Qt.PartiallyChecked)
        self.selected_molecules_checkbox.stateChanged.connect(self.on_selected_molecules_checkbox_state_change)
        self.selected_molecules_checkbox.setFocusPolicy(Qt.NoFocus)


        layout_bar.addWidget(QLabel('Selected'),0.1)
        layout_bar.addWidget(self.selected_molecules_checkbox, 0.15)

        self.molecule_index_field.returnPressed.connect(self.set_molecule_index_from_molecule_index_field)
        self.molecule_index_field.returnPressed.connect(self.deactivate_line_edit)


        layout.addLayout(layout_bar)
        layout.addWidget(self.canvas)

        self.setLayout(layout)
        # Create a placeholder widget to hold our toolbar and canvas.
        # widget = QWidget()
        # widget.setLayout(layout)
        # self.setCentralWidget(widget)

        self.dataset = dataset

        if show:
            self.show()

            app.exec_()

    def deactivate_line_edit(self):
        self.molecule_index_field.clearFocus()  # Clear the focus from the line edit

    @property
    def dataset(self):
        return self._dataset

    @dataset.setter
    def dataset(self, value):
        if value is not None and (hasattr(value, 'frame') or hasattr(value, 'time')):
            self._dataset = value
            self.canvas.init_plot_artists()
            self.set_selection()
            self.setDisabled(False)
        else:
            self._dataset = None
            self.setDisabled(True)
        self.molecule_index = 0

    @property
    def selection_state(self):
        return self._selection_state

    @selection_state.setter
    def selection_state(self, selection_state):
        self._selection_state = selection_state
        self.set_selection()

    def on_selected_molecules_checkbox_state_change(self, selection_state):
        self.selection_state = selection_state
        self.selected_molecules_checkbox.clearFocus()

    def set_selection(self):
        if self.selection_state == 0:
            self.dataset_molecule_indices_to_show = self.dataset.molecule.sel(molecule=~self.dataset.selected).values
        elif self.selection_state == 1:
            self.dataset_molecule_indices_to_show = self.dataset.molecule.values
        elif self.selection_state == 2:
            self.dataset_molecule_indices_to_show = self.dataset.molecule.sel(molecule=self.dataset.selected).values
        else:
            raise ValueError(f'Unknown selection_state {self.selection_state}')

    @property
    def molecule_index(self):
        return self._molecule_index

    @molecule_index.setter
    def molecule_index(self, molecule_index):
        self._molecule_index = molecule_index
        if self.dataset is not None and self.number_of_molecules_to_show > 0:
            self.molecule = self.dataset.isel(molecule=self.dataset_molecule_index)
        else:
            self.molecule = None
        self.molecule_index_field.setText(str(molecule_index))
        self.molecule_index_field.setFocusPolicy(Qt.ClickFocus)

    @property
    def dataset_molecule_index(self):
        return self.dataset_molecule_indices_to_show[self._molecule_index]

    @property
    def dataset_molecule_indices_to_show(self):
        return self._dataset_molecule_indices_to_show

    @dataset_molecule_indices_to_show.setter
    def dataset_molecule_indices_to_show(self, dataset_molecule_indices_to_show):
        self._dataset_molecule_indices_to_show = dataset_molecule_indices_to_show
        self.number_of_molecules_label.setText(f'{self.number_of_molecules_to_show}')
        self.molecule_index = 0

    @property
    def number_of_molecules_to_show(self):
        return len(self.dataset_molecule_indices_to_show)

    def set_molecule_index_from_molecule_index_field(self):
        self.molecule_index = int(self.molecule_index_field.text())

    def next_molecule(self):
        if (self.molecule_index+1) < self.number_of_molecules_to_show:
            self.molecule_index += 1

    def previous_molecule(self):
        if self.molecule_index > 0:
            self.molecule_index -= 1

    def update_current_molecule(self):
        self.molecule_index = self.molecule_index

    @property
    def molecule(self):
        return self.canvas.molecule

    @molecule.setter
    def molecule(self, molecule):
        self.canvas.molecule = molecule

    def keyPressEvent(self, e):
        key = e.key()
        if key == Qt.Key_Right: # Right arrow
            self.next_molecule()
        elif key == Qt.Key_Left: # Left arrow
            self.previous_molecule()
        elif key == Qt.Key_Space: # Spacebar
            self.dataset.selected[dict(molecule=self.dataset_molecule_index)] = ~self.dataset.selected[dict(molecule=self.dataset_molecule_index)]
            self.update_current_molecule()
        elif key == Qt.Key_S: # S
            self.canvas.save()

    # def selected_molecules_checkbox_state_changed(self, state):
    #     show_selected_mapping = {0: False, 1: None, 2: True}
    #     self.show_selected = show_selected_mapping[state]
    #     self.canvas.init_plot_artists()
    #     print('test')


class TracePlotCanvas(FigureCanvasQTAgg):
    # Kader om plot als geselecteerd
    # Autosave function
    def __init__(self, parent=None, width=14, height=7, dpi=100):
        self.figure = matplotlib.figure.Figure(figsize=(width, height), dpi=dpi, constrained_layout=True)  # , figsize=(2, 2))
        super().__init__(self.figure)
        self.parent_window = parent
        plot_variables = self.parent_window.plot_variables

        grid = self.figure.add_gridspec(len(plot_variables), 2, width_ratios=[10, 1]) #, height_ratios=(2, 7),
                         # left=0.1, right=0.9, bottom=0.1, top=0.9,
                         # wspace=0.05, hspace=0.05)

        self.plot_axes = {}
        self.histogram_axes = {}

        for i, plot_variable in enumerate(plot_variables):
            plot = self.figure.add_subplot(grid[i, 0])
            histogram = self.figure.add_subplot(grid[i, 1], sharey=plot)

            if i > 0:
                plot.sharex(self.plot_axes[plot_variables[0]])
                # histogram.sharex(self.histogram_axes[plot_variables[0]])

            plot.set_ylim(self.parent_window.ylims[i])
            plot.set_ylabel(plot_variable)

            histogram.get_yaxis().set_visible(False)

            self.plot_axes[plot_variable] = plot
            self.histogram_axes[plot_variable] = histogram


        # self.intensity_plot = self.figure.add_subplot(grid[0, 0])
        # self.FRET_plot = self.figure.add_subplot(grid[1, 0], sharex=self.intensity_plot)
        # self.intensity_histogram = self.figure.add_subplot(grid[0, 1], sharey=self.intensity_plot)
        # self.FRET_histogram = self.figure.add_subplot(grid[1, 1], sharex=self.intensity_histogram, sharey=self.FRET_plot)

        # self.figure = plt.Figure(dpi=dpi, figsize=(2,2))
        #
        # self.axis = self.figure.gca()

        #self.figure, self.axes = mpl.figure.Figure().subplots(2,1)



        self._molecule = None

        self.plot_artists = {}
        self.histogram_artists = {}


    def show_artists(self, show, draw=True):
        for artists in self.plot_artists.values():
            for artist in artists:
                artist.set_alpha(int(show))
        for artists in self.histogram_artists.values():
            for artist in artists:
                for bar in artist:
                    bar.set_alpha(int(show)*0.5)
        if draw:
            self.draw()

    def init_plot_artists(self):
        for i, plot_variable in enumerate(self.parent_window.plot_variables):
            data_array = self.parent_window.dataset[plot_variable]

            # For excluding nan values
            dims_wihtout_frame = set(data_array.dims).difference({'frame'})
            frame_not_nan = ~data_array.isnull().all(dim=dims_wihtout_frame)
            data_array = data_array.sel(frame=frame_not_nan)

            if 'time' in self.parent_window.dataset.coords.keys():
                x = data_array.time  # self.parent_window.dataset.time[frame_not_nan]
            else:
                x = data_array.frame  # self.parent_window.dataset.frame[frame_not_nan]
            self.plot_artists[plot_variable] = self.plot_axes[plot_variable].plot(x, data_array.sel(molecule=0).T)
            if i == 0:
                self.title_artist = self.plot_axes[plot_variable].set_title('')
            for j, plot_artist in enumerate(self.plot_artists[plot_variable]):
                plot_artist.set_color(self.parent_window.colours[i][j])
            # molecule.intensity.plot.line(x='frame', ax=self.plot_axes[plot_variable], color=self.parent_window.colours[i])
            self.histogram_artists[plot_variable] = self.histogram_axes[plot_variable].hist(data_array.sel(molecule=0).T,
                                                                                            bins=50,
                                                                                            orientation='horizontal',
                                                                                            # range=self.plot_axes[
                                                                                            #     plot_variable].get_ylim(),
                                                                                            range=self.parent_window.ylims[i],
                                                                                            color=
                                                                                            self.parent_window.colours[
                                                                                                i], alpha=0.5)[2]
            if not isinstance(self.histogram_artists[plot_variable], list):
                self.histogram_artists[plot_variable] = [self.histogram_artists[plot_variable]]

            if i == len(self.parent_window.plot_variables) - 1:
                if 'time' in self.parent_window.dataset.coords.keys():
                    self.plot_axes[plot_variable].set_xlabel(f'Time ({self.parent_window.dataset.time.units})')
                else:
                    self.plot_axes[plot_variable].set_xlabel('Frame')

        # self.artists += [self.intensity_plot.plot(g, c='g')]
        # self.artists += [self.intensity_plot.plot(r, c='r')]
        # self.artists += [self.FRET_plot.plot(e, c='b')]
        # self.artists += [[self.intensity_plot.set_title('test')]]
        # self.artists += [self.intensity_histogram.hist(g, bins=100, orientation='horizontal',
        #                                                range=self.intensity_plot.get_ylim(), color='g', alpha=0.5)[2]]
        # self.artists += [self.intensity_histogram.hist(r, bins=100, orientation='horizontal',
        #                                                range=self.intensity_plot.get_ylim(), color='r', alpha=0.5)[2]]
        # self.artists += [self.FRET_histogram.hist(e, bins=100, orientation='horizontal',
        #                                           range=self.FRET_plot.get_ylim(), color='b')[2]]

        # self.axes[1].plot(molecule.E(), animate=True)
        artists = [self.title_artist] + \
                  [a for b in self.plot_artists.values() for a in b] + \
                  [a for c in self.histogram_artists.values() for b in c for a in b]
        self.bm = BlitManager(self, artists)
        self.draw()

    @property
    def molecule(self):
        return self._molecule

    @molecule.setter
    def molecule(self, molecule):
        previous_molecule = self._molecule
        self._molecule = molecule

        if molecule is None and previous_molecule is not None:
            self.show_artists(False, draw=True)
            return
        elif molecule is None and previous_molecule is None:
            return
        elif molecule is not None and previous_molecule is None:
            self.show_artists(True, draw=False)


        self._molecule['file'] = self._molecule['file'].astype(str)

        # g = molecule.intensity.sel(channel=0).values
        # r = molecule.intensity.sel(channel=1).values
        # e = molecule.FRET.values

        if not self.plot_artists:
            self.init_plot_artists()

        # for axis in self.axes:
        #     axis.cla()

        for i, plot_variable in enumerate(self.parent_window.plot_variables):
            data = np.atleast_2d(molecule[plot_variable])

            # For excluding nan values (can go wrong when trace contains nans that are not present in all molecules)
            data = data[:, ~np.isnan(data).all(axis=0)]

            if self._molecule.selected.item():
                selection_string = ' | Selected'
            else:
                selection_string = ''

            self.title_artist.set_text(f'# {self.parent_window.molecule_index} of {len(self.parent_window.dataset.molecule)} | File: {molecule.file.values} | Molecule: {molecule.molecule_in_file.values}' + selection_string)#| Sequence: {molecule.sequence_name.values}')
            self.title_artist.set_text(
                f'File: {molecule.file.values} | Molecule: {molecule.molecule_in_file.values}' + selection_string)  # | Sequence: {molecule.sequence_name.values}')

            for j in range(len(data)):
                self.plot_artists[plot_variable][j].set_ydata(data[j])
                # TODO: When you shift the view, change the y positions of the bars to the new view, if possible. use set_y
                n, _ = np.histogram(data[j], 50, range=self.parent_window.ylims[i]) # range=self.plot_axes[plot_variable].get_ylim())
                for count, artist in zip(n, self.histogram_artists[plot_variable][j]):
                    artist.set_width(count)



        # self.artists[0][0].set_ydata(g)
        # self.artists[1][0].set_ydata(r)
        # self.artists[2][0].set_ydata(e)
        # self.artists[3][0].set_text(molecule.sequence_name.values)
        # n, _ = np.histogram(g, 100, range=self.intensity_plot.get_ylim())
        # for count, artist in zip(n, self.artists[4]):
        #     artist.set_width(count)
        # n, _ = np.histogram(r, 100, range=self.intensity_plot.get_ylim())
        # for count, artist in zip(n, self.artists[5]):
        #     artist.set_width(count)
        # n, _ = np.histogram(e, 100, range=self.FRET_plot.get_ylim())
        # for count, artist in zip(n, self.artists[6]):
        #     artist.set_width(count)
        #     #for count, rect in zip(n, bar_container.patches):
        # tell the blitting manager to do its thing
        self.bm.update()



        # self.axes[0].plot(molecule.intensity.T)
        # self.axes[1].plot(molecule.E())
        # self.canvas.draw()

        # empty_cell = (0, 0)
        #
        # donor_checkbox = wx.CheckBox(self, -1, label="Donor", name="Donor")
        # acceptor_checkbox = wx.CheckBox(self, -1, label="Acceptor", name="Acceptor")
        #
        # channel_sizer = wx.FlexGridSizer(2,2, gap=wx.Size(10,0))
        # channel_sizer.Add(wx.StaticText(self, label="Channels:"), 5, wx.EXPAND, 0)
        # channel_sizer.Add(donor_checkbox, 1, wx.EXPAND, 0)
        # channel_sizer.Add(empty_cell, 0, wx.EXPAND, 0)
        # channel_sizer.Add(acceptor_checkbox, 0, wx.EXPAND, 0)
        #
        # average_image_radio_button = wx.RadioButton(self, -1, label="Average image", name="Average image")
        # maximum_projection_radio_button = wx.RadioButton(self, -1, label="Maximum projection", name="Maximum projection")
        #
        # image_type_sizer = wx.FlexGridSizer(2,2, gap=wx.Size(10,0))
        # image_type_sizer.AddGrowableCol(0,1)
        # image_type_sizer.AddGrowableCol(1,3)
        # image_type_sizer.Add(wx.StaticText(self, label="Image type atestat:"), 0, wx.EXPAND, 0)
        # image_type_sizer.Add(average_image_radio_button, 0, wx.EXPAND, 0)
        # image_type_sizer.Add(empty_cell, 0, wx.EXPAND, 0)
        # image_type_sizer.Add(maximum_projection_radio_button, 0, wx.EXPAND, 0)
        # #
        # # image_type_sizer = wx.BoxSizer(wx.HORIZONTAL)
        # # image_type_sizer.Add(wx.StaticText(self, label="Image type:"), 0, wx.EXPAND | wx.ALL, 10)
        # # image_type_sizer.Add(image_type_combobox, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 10)
        #
        # self.sizer = wx.BoxSizer(wx.VERTICAL)
        # self.sizer.Add(channel_sizer, 0, wx.EXPAND | wx.ALL, 10)
        # self.sizer.Add(image_type_sizer, 0, wx.EXPAND | wx.ALL, 10)
        # self.sizer.Add(wx.Slider(self, -1, value=0, minValue=0, maxValue=100,
        #                          name='Minimum'), 0, wx.EXPAND | wx.ALL, 10)
        # self.sizer.Add(wx.Slider(self, -1, value=0, minValue=0, maxValue=100,
        #                          name='Maximum'), 0, wx.EXPAND | wx.ALL, 10)
        # self.sizer.Add(wx.Button(self, -1, "Button 1"), 0, wx.EXPAND | wx.ALL, 10)
        # self.sizer.Add(wx.Button(self, -1, "Button 2"), 0, wx.EXPAND | wx.ALL, 10)
        # self.SetSizer(self.sizer)

    def save(self):
        save_path = self.parent_window.save_path
        if save_path is not None:
            save_path.mkdir(parents=True, exist_ok=True)
            file_name = self.molecule.file.item().replace('\\' ,' - ')+f' - mol {self.molecule.molecule_in_file.item()}.png'
            file_path = save_path.joinpath(file_name)
            self.figure.savefig(file_path, bbox_inches='tight')
        else:
            raise ValueError('No save_path set')





#
#
#
# class TraceAnalysisFrame(QMainWindow):
#     def __init__(self, parent=None, dataset=None, title='Traces', plot_variables=['intensity', 'FRET'],
#                  ylims=[(0, 35000), (0, 1)], colours=[('g', 'r'), ('b')], save_path=None):
#         wx.Frame.__init__(self, parent, title=title, size=(1400, 700))
#         self.parent = parent
#         self.dataset = dataset
#         self.plot_variables = plot_variables
#         self.ylims = ylims
#         self.colours = colours
#
#         if save_path is None:
#             self.save_path = save_path
#         else:
#             self.save_path = Path(save_path)
#         #self.Bind(wx.EVT_CLOSE, self.OnClose)
#         self.trace_panel = TraceAnalysisPanel(parent=self)
#         # self.control_panel = ControlPanel(parent=self)
#         self.Bind(wx.EVT_CHAR_HOOK, self.OnNavigationKey)
#
#         self.molecule_index = 0
#
#         self.Show()
#
#     # @property
#     # def molecules(self):
#     #     return self._molecules
#     #
#     # @molecules.setter
#     # def molecules(self, molecules):
#     #     self._molecules = molecules
#
#
#     @property
#     def molecule_index(self):
#         return self._molecule_index
#
#     @molecule_index.setter
#     def molecule_index(self, molecule_index):
#         self._molecule_index = molecule_index
#         self.molecule = self.dataset.isel(molecule=self.molecule_index)
#
#     def next_molecule(self):
#         if (self.molecule_index+1) < len(self.dataset.molecule):
#             self.molecule_index += 1
#
#     def previous_molecule(self):
#         if self.molecule_index > 0:
#             self.molecule_index -= 1
#
#     def update_current_molecule(self):
#         self.molecule_index = self.molecule_index
#
#     @property
#     def molecule(self):
#         return self.panel.molecule
#
#     @molecule.setter
#     def molecule(self, molecule):
#         self.trace_panel.molecule = molecule
#
#     def OnNavigationKey(self, event):
#         key_code = event.GetKeyCode()
#         print(key_code)
#         if key_code == 316: # Right arrow
#             self.next_molecule()
#         elif key_code == 314: # Left arrow
#             self.previous_molecule()
#         elif key_code == 32: # Spacebar
#             self.dataset.selected[dict(molecule=self.molecule_index)] = ~self.dataset.selected[dict(molecule=1)]
#             self.update_current_molecule()
#         elif key_code == 83: # S
#             self.trace_panel.save()

# class ControlPanel(wx.Panel):
# To file/molecule
# Selected button changes colour (spacebar)
# Y-axis limits
# Classification technique - change classification pannel based on specific technique

# class Threshold_classification_panel

# class HMM_classification_panel


class BlitManager:
    def __init__(self, canvas, animated_artists=()):
        """
        Parameters
        ----------
        canvas : FigureCanvasAgg
            The canvas to work with, this only works for sub-classes of the Agg
            canvas which have the `~FigureCanvasAgg.copy_from_bbox` and
            `~FigureCanvasAgg.restore_region` methods.

        animated_artists : Iterable[Artist]
            List of the artists to manage
        """
        self.canvas = canvas
        self._bg = None
        self._artists = []

        for a in animated_artists:
            self.add_artist(a)
        # grab the background on every draw
        self.cid = canvas.mpl_connect("draw_event", self.on_draw)

    def on_draw(self, event):
        """Callback to register with 'draw_event'."""
        cv = self.canvas
        if event is not None:
            if event.canvas != cv:
                raise RuntimeError
        self._bg = cv.copy_from_bbox(cv.figure.bbox)
        self._draw_animated()

    def add_artist(self, art):
        """
        Add an artist to be managed.

        Parameters
        ----------
        art : Artist

            The artist to be added.  Will be set to 'animated' (just
            to be safe).  *art* must be in the figure associated with
            the canvas this class is managing.

        """
        if art.figure != self.canvas.figure:
            raise RuntimeError
        art.set_animated(True)
        self._artists.append(art)

    def _draw_animated(self):
        """Draw all of the animated artists."""
        fig = self.canvas.figure
        for a in self._artists:
            fig.draw_artist(a)

    def update(self):
        """Update the screen with animated artists."""
        cv = self.canvas
        fig = cv.figure
        # paranoia in case we missed the draw event,
        if self._bg is None:
            self.on_draw(None)
        else:
            # restore the background
            cv.restore_region(self._bg)
            # draw all of the animated artists
            self._draw_animated()
            # update the GUI state
            cv.blit(fig.bbox)
        # let the GUI event loop process anything it has to do
        # cv.flush_events()

# class MainWindow(wx.Frame):
#    def __init__(self, parent, title):
#        wx.Frame.__init__(self, parent, title=title, size=(300, 700))
#        self.parent = parent
#        self.panel = TraceAnalysisPanel(parent=self)
#        # self.Bind(wx.EVT_CLOSE, self.OnClose)
#        self.Show()

if __name__ == "__main__":

    # # Check whether there is already a running QApplication (e.g., if running
    # # from an IDE).
    # qapp = QtWidgets.QApplication.instance()
    # if not qapp:
    #     qapp = QtWidgets.QApplication(sys.argv)
    #
    # app = ApplicationWindow()
    # app.show()
    # app.activateWindow()
    # app.raise_()
    # qapp.exec_()


    import papylio as pp
    import os, sys
    mapping_path = Path(os.getcwd()).joinpath('papylio').joinpath('mapping')
    sys.path.append(mapping_path)
    print(sys.path)
    from papylio.experiment import Experiment
    exp = Experiment(r'D:\SURFdrive\Promotie\Code\Python\traceAnalysis\twoColourExampleData\20141017 - Holliday junction - Copy')
    ds = exp.files[0].dataset

    from PySide2.QtWidgets import QApplication

    app = QApplication(sys.argv)
    frame = TracePlotWindow(ds)
        #, "Sample editor", plot_variables=['intensity', 'FRET'],  # 'classification'],
        #          ylims=[(0, 1000), (0, 1), (-1,2)], colours=[('g', 'r'), ('b'), ('k')])

    app.exec_()

    # # exp = pp.Experiment(r'D:\20200918 - Test data\Single-molecule data small')
    # #exp = pp.Experiment(r'P:\SURFdrive\Promotie\Data\Test data')
    # # exp = pp.Experiment(r'/Users/ivoseverins/SURFdrive/Promotie/Data/Test data')
    # # print(exp.files)
    # # m = exp.files[1].molecules[0]
    # # print(exp.files[2])
    # import xarray as xr
    # #file_paths = [p for p in exp.nc_file_paths if '561' in str(p)]
    # file_paths = [exp.nc_file_paths[0]]
    # with xr.open_mfdataset(file_paths, concat_dim='molecule', combine='nested') as ds:
    #     # ds_sel = ds.sel(molecule=ds.sequence_name=='HJ7_G')# .reset_index('molecule', drop=True) # HJ1_WT, HJ7_G116T
    #     app = wx.App(False)
    #     # app = wit.InspectableApp()
    #     frame = TraceAnalysisFrame(None, ds, "Sample editor", plot_variables=['intensity', 'FRET'], #'classification'],
    #              ylims=[(0, 1000), (0, 1), (-1,2)], colours=[('g', 'r'), ('b'), ('k')])
    #     # frame.molecules = exp.files[1].molecules
    #     print('test')
    #     import wx.lib.inspection
    #     wx.lib.inspection.InspectionTool().Show()
    #     app.MainLoop()





# Add time to existing .nc file
# for file in exp.files:
#     with xr.open_dataset(file.absoluteFilePath.with_suffix('.nc')) as ds:
#         i = ds.intensity.load()
#     test = i.assign_coords(time=file.movie.time)
#     test.to_netcdf(file.absoluteFilePath.with_suffix('.nc'), engine='h5netcdf', mode='a')


#
# from matplotlib import use
# use('TkAgg')
#
# import papylio as pp
# exp = pp.Experiment(r'D:\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
# #exp = pp.Experiment(r'J:\Ivo\20200221 - Magnetic tweezers setup (Old)\Data')
# # exp.files[-2].perform_mapping()
# # exp.files[-2].mapping.show_mapping_transformation()

# class B:
#     def __init__(self):
#         print('Badd')
#         super().__init__()
#
#
#
# class A:
#     def __init__(self):
#         print('A')
#
#
# def test(c):
#     return type(c.__name__, (c,B),{})
#
#
# @test
# class Bo(A):
#     def __init__(self):
#         print('Bo')
#         super().__init__()



# class B:
#     def __init__(self):
#         print('Badd')
#         super().__init__()
#
# # class PluginMetaClass(type):
# #     def __new__(cls, clsname, bases, attrs):
# #         bases_base = tuple(base for base in bases if not base.__name__ is clsname)
# #         attrs.pop('__qualname__')
# #         cls_base = type(clsname+'_base', bases_base, attrs)
# #         bases_main = tuple(base for base in bases if base.__name__ is clsname) + (cls_base,)
# #         return super().__new__(cls, clsname, bases_main, {})
# class PluginMetaClass(type):
#     def __new__(cls, clsname, bases_base, attrs):
#         # bases_base = tuple(base for base in bases if not base.__name__ is clsname)
#         attrs_base = attrs.copy()
#         attrs_base.pop('__qualname__')
#         #attrs_base.pop('__module__')
#         #attrs_base.pop('__classcell__')
#         cls_base = super().__new__(cls, clsname, bases_base, attrs_base)
#         #cls_base = type(clsname, bases_base, attrs)
#         added_bases = (B,)
#         bases_main = added_bases + (cls_base,)
#         test = super().__new__(cls, clsname+'main', bases_main,{})
#         print('test')
#         return test
#
# class A:
#     def __init__(self):
#         print('A')
#
# class Bo(A, metaclass=PluginMetaClass):
#     def __init__(self):
#         print('Bo')
#         super().__init__()
#



# exp = pp.Experiment(r'P:\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
# # exp = pp.Experiment(r'D:\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
# exp.files[-1].use_mapping_for_all_files()



# def add_class_to_class(base_class):
#     def add_class_to_class_decorator(added_class):
#         base_class.__bases__ += (added_class,)
#     return add_class_to_class_decorator
#
# @add_class_to_class(pp.File)
# class ExperimentPlugIn():
#     def test(self):
#         print(self.name)



# exp.files[0].find_coordinates()
#
#
# # #exp = pp.Experiment(r'D:\ivoseverins\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20191209 - Single-molecule setup (TIR-I)')
# # exp.files[0].perform_mapping(transformation_type='nonlinear')
# #
# import matplotlib.pyplot as plt
# figure = plt.figure()
# #exp.files[0].show_average_image(figure=figure)
# plt.imshow(exp.files[0].movie.maximum_projection_image)
# exp.files[0].show_coordinates(figure=figure)
# #exp.files[0].mapping.show_mapping_transformation(figure=figure)



# exp.files[-1].use_mapping_for_all_files()

from papylio.plotting import histogram
# exp.files[7].histogram(bins = 100, molecule_averaging=True, export=True)
# exp.histogram(bins = 100, molecule_averaging=True, export=True)
#
# import sys
# #sys.path.append(r'D:\ivoseverins\SURFdrive\Promotie\Code\Python\fastqAnalysis')
# sys.path.append(r'D:\SURFdrive\Promotie\Code\Python\fastqAnalysis')
#
# from papylio.traceAnalysisCode import Experiment
# from fastqAnalysis import FastqData
#
# from pathlib import Path # For efficient path manipulation
#
# path = Path(r'G:\Ivo\20190918 - Sequencer (MiSeq)\Analysis')
# #path = 'D:\\ivoseverins\\Desktop\\Sequencing data\\20180705\\'
# #path = 'C:\\Users\\Ivo Severins\\Desktop\\Sequencing data\\20180705\\'
# fileName = r'One_S1_L001_R1_001.fastq'
#
#
# data = FastqData(path.joinpath(fileName))
#
# data.selection(sequence = 'AA')
#
# data.matches_per_tile(sequence = 'TATCTGTATAATGAGAAATATGGAGTACAATTTTTTTTTTTTTTTTTTTT')









#import wx
#
#
#class OtherFrame(wx.Frame):
#    """
#    Class used for creating frames other than the main one
#    """
#
#    def __init__(self, title, parent=None):
#        wx.Frame.__init__(self, parent=parent, title=title)
#        self.Show()
#
#
#class MyPanel(wx.Panel):
#
#    def __init__(self, parent):
#        wx.Panel.__init__(self, parent)
#
#        btn = wx.Button(self, label='Create New Frame')
#        btn.Bind(wx.EVT_BUTTON, self.on_new_frame)
#        self.frame_number = 1
#
#    def on_new_frame(self, event):
#        title = 'SubFrame {}'.format(self.frame_number)
#        frame = OtherFrame(title=title)
#        self.frame_number += 1
#
#
#class MainFrame(wx.Frame):
#
#    def __init__(self):
#        wx.Frame.__init__(self, None, title='Main Frame', size=(800, 600))
#        panel = MyPanel(self)
#        self.Show()
#
#
#if __name__ == '__main__':
#    app = wx.App(False)
#    frame = MainFrame()
#    app.MainLoop()


# #!/usr/bin/env python
# import wx
# import wx.dataview
# import wx.lib.agw.aui as aui
# import os
#
# import wx.lib.agw.customtreectrl as CT
# #from traceAnalysisCode import Experiment
# import wx.lib.agw.hypertreelist as HTL
#
#
# import matplotlib as mpl
# from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
# from matplotlib.backends.backend_wxagg import NavigationToolbar2WxAgg as NavigationToolbar
#
# from matplotlib import use
# use('WXAgg')
# from matplotlib import pyplot as plt
# #import matplotlib.pyplot as plt
#
#
#
# class MyFrame(wx.Frame):
#     """ We simply derive a new class of Frame. """
#     def __init__(self, parent, title):
#         wx.Frame.__init__(self, parent, title=title, size=(400,400))
#         tree_list = HTL.HyperTreeList(self)
#
#         tree_list.AddColumn("First column")
#
#         root = tree_list.AddRoot("Root")
#
#         parent = tree_list.AppendItem(root, "First child")
#         child = tree_list.AppendItem(parent, "First Grandchild")
#
#         tree_list.AppendItem(root, "Second child", ct_type=1)
#         self.Show(True)
#
# app = wx.App(False)
# frame = MyFrame(None, 'Small editor')
# app.MainLoop()


import numpy as np

def generate_sequences(base_composition):
    """

    Parameters
    ----------
    base_composition : list of str
        A list with the length of the sequence, where each entry indicates the possible bases at that position.
        pairI where I is an integer indicates that the base at the position of pairI should basepair with position I.
        E.g. ['N', 'AC', 'GC', 'ACTG', 'pair1']

    Returns
    -------
    sequences : numpy.array
        Array of sequences
    """

    base_composition_2 = []
    for bases in base_composition:
        if bases == 'N':
            bases = 'ACTG'
        elif bases.startswith('pair'):
            bases = '-'

        base_composition_2.append(list(bases))

    sequences = np.array(np.meshgrid(*base_composition_2)).T.reshape(-1, len(base_composition_2))

    for i, bases in enumerate(base_composition):
        if bases.startswith('pair'):
            sequences[:,i] = convert_bases(sequences[:,int(bases[4:])], conversion='basepair')
    return sequences.view(f'U{len(base_composition_2)}').squeeze()

def convert_bases(bases, conversion='basepair', **conversion_dict):
    if conversion == 'basepair':
        conversion_dict = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}

    return np.array([conversion_dict[base] for base in bases])


    # DNA_bases = ['A','T','C','G']
    # complementary_DNA_bases =
    #
    # condlist = [bases == base for base in
    #
    # np.select(condlist, choicelist, 55)


import re
import sys
import itertools
import warnings
import tqdm
import re
from numba import njit
from pathlib import Path
import pandas as pd
import tifffile
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import xarray as xr
import scipy.ndimage
from skimage.transform import AffineTransform

# from papylio.movie.background_correction import rollingball
from papylio.movie.background_correction import determine_temporal_background_correction, \
    determine_spatial_background_correction, determine_single_value_background_correction # remove_background, get_threshold
from papylio.timer import Timer

class Illumination:
    def __init__(self, name, short_name='', other_names=[]):
        # self.movie = movie
        self.name = name
        self.short_name = short_name
        self.other_names = other_names

    def __repr__(self):
        return (f'{self.__class__.__name__}({self.name})')

    @property
    def names(self):
        return [self.index, self.name, self.short_name] + self.other_names

    @property
    def index(self):
        try:
            return Movie.illuminations.index(self)
        except:
            pass


class Movie:
    @classmethod
    def type_dict(cls):
        # It is important to import all movie files to recognize them by subclasses.
        # Perhaps we can make this more elegant in some way.
        from papylio.movie.sifx import SifxMovie
        from papylio.movie.pma import PmaMovie
        from papylio.movie.tif import TifMovie
        from papylio.movie.nd2 import ND2Movie
        from papylio.movie.nsk import NskMovie
        from papylio.movie.binary import BinaryMovie
        return {extension: subclass for subclass in cls.__subclasses__() for extension in subclass.extensions}

    illuminations = [Illumination('green', 'g'), Illumination('red', 'r')]

    @classmethod
    def get_illumination_from_name(cls, illumination_name):
        """Get the channel index belonging to a specific channel (name)
        If

        Parameters
        ----------
        channel : str or int
            The name or number of a channel

        Returns
        -------
        i: int
            The index of the channel to which the channel name belongs

        """
        for illumination in cls.illuminations:
            if illumination_name in illumination.names or illumination_name == illumination:
                return illumination
        else:
            raise ValueError('Illumination name not found')

    @classmethod
    def get_illuminations_from_names(cls, illumination_names):
        """Get the channel index belonging to a specific channel (name)
        If

        Parameters
        ----------
        channel : str or int
            The name or number of a channel

        Returns
        -------
        i: int
            The index of the channel to which the channel name belongs

        """
        if illumination_names in [None, 'all']:
            return cls.illuminations

        if not isinstance(illumination_names, list):
            illumination_names = [illumination_names]

        return [cls.get_illumination_from_name(illumination_name) for illumination_name in illumination_names]

    @classmethod
    def get_illumination_indices_from_names(cls, illumination_names):
        illuminations = cls.get_illuminations_from_names(illumination_names)
        return [illumination.index for illumination in illuminations]

    @classmethod
    def image_info_from_filename(cls, filename):
        image_info = {}

        fov_index_result = re.search('(?<=_fov)\d*(?=[_.])', filename)
        if fov_index_result is not None:
            image_info['fov_index'] = int(fov_index_result.group())

        if '_ave' in filename:
            image_info['projection_type'] = 'average'
        elif '_max' in filename:
            image_info['projection_type'] = 'maximum'

        frame_start = re.search('(?<=_f)\d*(?=[-])', filename)
        if frame_start is not None:
            frame_end = re.search(f'(?<=_f{frame_start.group()}-)\d*(?=[-_.])', filename)
            frame_interval = re.search(f'(?<=_f{frame_start.group()}-{frame_end.group()}-)\d*(?=[_.])', filename)
            if frame_end is not None:
                frame_range = (int(frame_start.group()), int(frame_end.group()))
            else:
                raise ValueError('Invalid filename')
            if frame_interval is not None:
                frame_range += (int(frame_interval.group()),)
            image_info['frame_range'] = frame_range

        illumination_result = re.search('(?<=_i)\d*(?=[_.])', filename)
        if illumination_result is None:
            image_info['illumination_index'] = None  # list(self.illumination_indices.values)
        else:
            image_info['illumination_index'] = int(illumination_result.group())

        # channel_result = re.search('(?<=_c)\d*(?=[_.])', filename)
        # if channel_result is None:
        #     image_info['channel_indices'] = list(self.channel_indices.values)
        # else:
        #     image_info['channel_indices'] = int(channel_result.group())

        # fov_index = re.search('(?<=_fov)\d*(?=[_.])', filename)
        # if fov_index is not None:
        #     fov_index = int(fov_index)
        #     image_info['fov_index'] = fov_index

        illumination_result = re.search('_raw', filename)
        if illumination_result is None:
            image_info['apply_corrections'] = True
        else:
            image_info['apply_corrections'] = False

        return image_info

    @classmethod
    def image_info_to_filename(cls, filename, fov_index=None, projection_type=None, frame_range=None,
                               illumination=None, apply_corrections=None):
        # if 'fov_info' in self.__dict__.keys() and self.fov_info: # Or hasattr(self, 'fov_info')
        if fov_index is not None:
            # filename += f'_fov{self.fov_info["fov_chosen"]:03d}'
            filename += f'_fov{fov_index:03d}'

        if projection_type is not None:
            filename += '_' + projection_type[:3]

        if frame_range is not None:
            filename += str(range(*frame_range)).replace('range(', '_f').replace(', ', '-').replace(')', '')

        if illumination is not None:  # and self.number_of_illuminations_in_movie > 1:
            illumination_index = cls.get_illumination_from_name(illumination).index
            filename += f'_i{illumination_index}'

        # if channel is not None:  # and self.number_of_illuminations_in_movie > 1:
        #     channel_index = cls.get_channel_from_name(channel).index
        #     filename += f'_i{channel_index}'

        if apply_corrections is False:
            filename += '_raw'

        return filename

    def __new__(cls, filepath, rot90=0):
        if cls is Movie:
            extension = Path(filepath).suffix.lower()
            try:
                return object.__new__(cls.type_dict()[extension])
            except KeyError:
                raise NotImplementedError('Filetype not supported')
        else:
            return object.__new__(cls)

    def __getnewargs__(self):
        return (self.filepath, self.rot90)

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('file', None)
        return d

    def __setstate__(self, dict):
        self.__dict__.update(dict)

    def __init__(self, filepath, rot90=0):  # , **kwargs):
        self.filepath = Path(filepath)
        self._with_counter = 0
        self.fov_index = None
        # self.filepaths = [Path(filepath) for filepath in filepaths] # For implementing multiple files, e.g. two channels over two files
        self.is_mapping_movie = False

        self.rot90 = rot90
        # self.correct_images = False

        self.chunk_size = 100
        self.use_dask = False

        self._data_type = np.dtype(np.uint16)
        self.intensity_range = (np.iinfo(self.data_type).min, np.iinfo(self.data_type).max)

        if not self.filepath.suffix == '.sifx':
            self.writepath = self.filepath.parent
            self.name = self.filepath.with_suffix('').name

        self._time = None

        self.channels = [Channel(self, 'green', 'g', other_names=['donor', 'd']),
                         Channel(self, 'red', 'r', other_names=['acceptor', 'a'])]
        self.channel_arrangement = [[[0, 1]]]
        # [[[0,1]]] # First level: frames, second level: y within frame, third level: x within frame
        # self.channel_arrangement = xr.DataArray([[[0,1]]], dims=('frame','y','x'))

        self.illumination_arrangement = [0]  # First level: frames, second level: illumination
        # self.illumination_arrangement = xr.DataArray([[True, False]], dims=('frame', 'illumination'), coords={'illumination': [0,1]}) # TODO: np.array([0]) >> list of list It would be good to have a default illumination_arrangement of np.array([0]), i.e. illumination 0 all the time?
        self._illumination_index_per_frame = None

        # self._darkfield_correction = None
        # self._flatfield_correction = None
        # self._general_background_correction = None
        # self._spatial_background_correction = None
        # self._temporal_background_correction = None
        # self._temporal_illumination_correction = None

        self._common_corrections = xr.Dataset()

        # self.load_corrections()

        self.header_is_read = False

    def __enter__(self):
        if self._with_counter == 0:
            self.open()
            # print('open')
        self._with_counter += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._with_counter -= 1
        if self._with_counter == 0:
            self.close()
            # print('close')

    def __repr__(self):
        return (f'{self.__class__.__name__}({str(self.filepath)})')

    def __getattr__(self, item):
        # if '_initialized' in self.__dict__ and not self.header_is_read:
        # if item != 'header_is_read' and not self.header_is_read:
        # if item == '_with_counter':
        #     raise ValueError()
        # print(item)

        if 'header_is_read' in self.__dict__.keys() and not self.header_is_read:
            # print(item+'2')
            self.read_header()
            return getattr(self, item)
        else:
            raise AttributeError(f'Attribute {item} not found')
        # return super().__getattribute__(item)

    @property
    def pixels_per_frame(self):
        return self.width * self.height

    @property
    def bitdepth(self):
        return self.data_type.itemsize * 8  # 8 bits in a byte

    @property
    def bytes_per_frame(self):
        return self.data_type.itemsize * self.pixels_per_frame

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, value):
        self._time = value

    # @property
    # def channel_grid(self):
    #     """ numpy.array : number of channels in the horizontal and vertical dimension
    #
    #     Setting the channel_grid variable will assume equally spaced channels
    #     """
    #     return self._channel_grid
    #
    # @channel_grid.setter
    # def channel_grid(self, channel_grid):
    #     channel_grid = np.array(channel_grid)
    #     # Possibly support multiple cameras by adding a third dimension
    #     if len(channel_grid) == 2 and np.all(np.array(channel_grid) > 0):
    #         self._channel_grid = channel_grid
    #         self._number_of_channels = np.product(channel_grid)
    @property
    def number_of_illuminations(self):
        """ int : number of channels in the movie

        Setting the number of channels will divide the image horizontally in equally spaced channels.
        """
        return len(self.illuminations)

    @property
    def number_of_illuminations_in_movie(self):
        """ int : number of channels in the movie

        Setting the number of channels will divide the image horizontally in equally spaced channels.
        """
        return len(self.illumination_indices_in_movie)

    @property
    def number_of_channels(self):
        """ int : number of channels in the movie

        Setting the number of channels will divide the image horizontally in equally spaced channels.
        """
        return len(self.channels)

    @property
    def data_type(self):
        return self._data_type

    @data_type.setter
    def data_type(self, data_type):
        self._data_type = data_type
        self.intensity_range = (np.iinfo(self.data_type).min, np.iinfo(self.data_type).max)

    @property
    def frame_indices(self):
        return xr.DataArray(np.arange(self.number_of_frames), dims='frame')

    @property
    def channel_arrangement(self):
        return self._channel_arrangement

    @channel_arrangement.setter
    def channel_arrangement(self, channel_arrangement):
        self._channel_arrangement = np.array(channel_arrangement)

    @property
    def channel_indices(self):
        return xr.DataArray(self.channel_arrangement.flatten(), dims='channel')

    # @property
    # def channel_indices_per_image(self):
    #     if self._channel_indices_per_frame is None:
    #         # frame_indices = self.frame_indices
    #         # illumination_indices = self.illumination_indices
    #         # self._illumination_index_per_frame = xr.DataArray(
    #         #     np.resize(self.illumination_arrangement, (len(frame_indices), len(illumination_indices))),
    #         #     dims=('frame', 'illumination'),
    #         #     coords={'frame': frame_indices, 'illumination': illumination_indices})
    #         channel_indices_flattened = self.channel_arrangement.reshape(len(self.channel_arrangement), -1)
    #         self._channel_indices_per_frame= xr.DataArray(
    #             np.resize(channel_indices_flattened, (self.number_of_frames, len(channel_indices_flattened[0]))),
    #             dims=('frame', 'channel'),
    #             coords={'frame': self.frame_indices}).stack(image=('frame','channel'))
    #         #TODO: Add name to other indices or remove this name
    #     return self._illumination_index_per_frame

    @property
    def number_of_channels_per_frame(self):
        return np.product(self.channel_arrangement.shape[1:])

    @property
    def illumination_arrangement(self):
        return self._illumination_arrangement

    @illumination_arrangement.setter
    def illumination_arrangement(self, illumination_arrangement):
        self._illumination_arrangement = np.array(illumination_arrangement)
        self._illumination_index_per_frame = None

    @property
    def illumination_indices(self):
        return xr.DataArray([illumination.index for illumination in self.illuminations], dims='illumination')

    @property
    def illumination_index_per_frame(self):
        if self._illumination_arrangement is not None and self._illumination_index_per_frame is None:
            # frame_indices = self.frame_indices
            # illumination_indices = self.illumination_indices
            # self._illumination_index_per_frame = xr.DataArray(
            #     np.resize(self.illumination_arrangement, (len(frame_indices), len(illumination_indices))),
            #     dims=('frame', 'illumination'),
            #     coords={'frame': frame_indices, 'illumination': illumination_indices})
            self._illumination_index_per_frame = xr.DataArray(
                np.resize(self.illumination_arrangement, (self.number_of_frames)),
                dims=('frame'),
                coords={'frame': self.frame_indices}, name='illumination')
            # TODO: Add name to other indices or remove this name
        return self._illumination_index_per_frame

    @illumination_index_per_frame.setter
    def illumination_index_per_frame(self, illumination_index_per_frame):
        self._illumination_index_per_frame = illumination_index_per_frame
        self._illumination_arrangement = None

    @property
    def illumination_indices_in_movie(self):
        return np.unique(self.illumination_index_per_frame)

    # @property
    # def image_indices(self):
    #     # index = pd.MultiIndex.from_arrays([*self.image_indices_from_frame_indices(self.frame_indices).T],
    #     #                                   names=('frame', 'illumination', 'channel'))
    #     # return xr.DataArray(index, dims='image')
    #     return self.image_indices_from_frame_indices(xarray=True)

    # def image_indices_from_frame_indices(self, frame_indices=None, xarray=False):
    #     if frame_indices is None:
    #         frame_indices = self.frame_indices
    #     if isinstance(frame_indices, xr.DataArray):
    #         frame_indices = frame_indices.values
    #     # return self.image_indices.sel(image=self.image_indices.frame.isin(frame_indices))
    #     image_frame_indices = np.repeat(frame_indices, self.number_of_channels_per_frame)
    #     image_illumination_indices = np.repeat(self.illumination_index_per_frame.values[frame_indices],
    #                                            self.number_of_channels_per_frame)
    #     image_channel_indices = np.resize(self.channel_indices.values, len(image_frame_indices))
    #
    #     image_indices = np.vstack([image_frame_indices, image_illumination_indices, image_channel_indices]).T
    #     if xarray:
    #         image_indices = self.image_indices_to_xarray(image_indices)
    #
    #     return image_indices

    # def image_indices_to_xarray(self, image_indices):
    #     index = pd.MultiIndex.from_arrays([*image_indices.T], names=('frame', 'illumination', 'channel'))
    #     return xr.DataArray(index, dims='image')

    # TODO: remove this
    # def create_frame_info(self):
    #     # TODO: Use xarray instead of pandas
    #     # Perhaps store time, illumination and channel separately
    #     # files = [0] # For implementing multiple files
    #     frames = range(self.number_of_frames)
    #
    #     index = pd.Index(data=frames, name='frame')
    #     frame_info = pd.DataFrame(index=index, columns=['time', 'illumination', 'channel'])
    #     # self.frame_info['file'] = len(self.frame_info) * [list(range(2))] # For implementing multiple files
    #     # self.frame_info = self.frame_info.explode('file') # For implementing multiple files
    #     frame_info['time'] = frame_info.index.to_frame()['frame'].values
    #     if self.illumination_arrangement is not None:
    #         if len(self.illumination_arrangement)>1:
    #             frame_info['illumination'] = self.illumination_arrangement.tolist() * (self.number_of_frames // self.illumination_arrangement.shape[0])
    #         else:
    #             frame_info['illumination'] = [0] * self.number_of_frames
    #     else:
    #         frame_info['illumination'] = [0] * self.number_of_frames
    #     frame_info['channel'] = self.channel_arrangement.tolist() * (self.number_of_frames // self.channel_arrangement.shape[0])
    #
    #     frame_info = frame_info.explode('channel').explode('channel')
    #
    #     categorical_columns = ['illumination', 'channel']
    #     frame_info[categorical_columns] = frame_info[categorical_columns].astype('category')
    #
    #     self.frame_info = frame_info

    @property
    def pixel_to_stage_coordinates_transformation(self):
        pixels_to_um = AffineTransform(scale=self.pixel_size)
        pixels_um_to_stage_coordinates_um = AffineTransform(translation=np.flip(self.stage_coordinates))
        pixels_to_stage_coordinates_um = pixels_to_um + pixels_um_to_stage_coordinates_um
        return pixels_to_stage_coordinates_um

    @property
    def width_metric(self):
        return self.width * self.pixel_size[0]

    @property
    def height_metric(self):
        return self.height * self.pixel_size[1]

    @property
    def boundaries_metric(self):
        # #         Formatted as two coordinates, with the lowest and highest x and y values respectively
        horizontal_boundaries = np.array([0, self.width_metric])
        vertical_boundaries = np.array([0, self.height_metric])
        return np.vstack([horizontal_boundaries, vertical_boundaries]).T

    @property
    def boundaries_stage(self):
        return self.pixel_to_stage_coordinates_transformation(self.channels[0].boundaries)

    def read_header(self):
        self._read_header()
        if not (self.rot90 % 2 == 0):
            width = self.width
            height = self.height
            self.width = height
            self.height = width

        self.header_is_read = True

    def read_frame(self, frame_index, **kwargs):
        return self.read_frames([frame_index], **kwargs).squeeze(axis=0)

    def read_frames(self, frame_indices=None, apply_corrections=True, xarray=True, flatten_channels=False):
        if frame_indices is None:
            frame_indices = self.frame_indices.values

        frames = self._read_frames(frame_indices)
        # frames = xr.DataArray(frames, dims=('frame', 'y', 'x'))
        frames = np.rot90(frames, self.rot90, axes=(1, 2))

        if len(self.channel_arrangement) > 1:
            raise NotImplementedError('Channel arrangement where frames indicated different channels not implemented')
            # Perhaps remove the outermost layer from channel_configuration
            # Or add this to separate and flatten channels

        frames = self.separate_channels(frames)
        # frames = np.stack([channel.crop_images(images) for channel in self.channels]

        if apply_corrections:  # and self.correct_images
            frames = self.apply_corrections(frames, frame_indices)

        if xarray:
            frames = self.frames_to_xarray_dataarray(frames, frame_indices)

        if flatten_channels:
            frames = self.flatten_channels(frames)

        return frames

    @property
    def channel_rows(self):
        return len(self.channel_arrangement[0])

    @property
    def channel_columns(self):
        return len(self.channel_arrangement[0, 0])

    def separate_channels(self, frames):
        # if frames.ndim == 2:
        #     frames = frames[None, :, :]
        # return expand_axes(frames, (channel_rows, channel_columns), from_axes=(1, 2))
        # return expand_axes(frames, (channel_rows, channel_columns), from_axes=(1, 2), to_axes=(0, 0))
        return xr.apply_ufunc(
            expand_axes, frames, input_core_dims=[['y', 'x']], output_core_dims=[['channel', 'y', 'x']],
            exclude_dims=set(['y', 'x']),
            kwargs={"expand_into": (self.channel_rows, self.channel_columns), "from_axes": (-2, -1),
                    "to_axes": (frames.ndim,) * 2, "new_axes_positions": [-3]}
        )
        # return xr.apply_ufunc(
        #     expand_axes, frames, input_core_dims=[['image', 'y', 'x'][-frames.ndim:]], output_core_dims=[['image', 'y', 'x']],
        #     exclude_dims=set(['image', 'y', 'x']),
        #     kwargs={"expand_into": (channel_rows, channel_columns), "from_axes": (-2, -1), "to_axes": (-3, -3)}
        # )

        # frames = frames.transpose('frame', 'y', 'x', ...)
        # new = split_along_axes(frames.values, (channel_rows, channel_columns), from_axes=(1, 2))
        # return xr.DataArray(new, dims=['frame','y','x','channel'])

    def flatten_channels(self, frames):
        # return split_along_axes(frames, (channel_rows, channel_columns), from_axes=(1, 2), inverse=True)
        # return split_along_axes(frames, (channel_rows, channel_columns), from_axes=(1, 2), inverse=True)
        # if frames.shape[-3]//channel_columns//channel_rows == 1:
        #     output_core_dims = [['y', 'x']]
        # else:
        #     output_core_dims = [['image', 'y', 'x']]

        return xr.apply_ufunc(
            expand_axes, frames, input_core_dims=[['channel', 'y', 'x']], output_core_dims=[['y', 'x']],
            exclude_dims=set(['x', 'y']),
            kwargs={"expand_into": (self.channel_rows, self.channel_columns), "from_axes": (-2, -1),
                    "to_axes": (-3, -3),
                    "inverse": True, "squeeze": True}
        )

    def frames_to_xarray_dataarray(self, frames, frame_indices):
        frames = xr.DataArray(frames,
                              dims=('frame', 'channel', 'y', 'x'),
                              coords={'frame': frame_indices,
                                      'illumination': self.illumination_index_per_frame[frame_indices],
                                      'channel': self.channel_indices})

        if self.time is not None:
            frames = frames.assign_coords(time=self.time[frames.frame])

        return frames

    def get_channel(self, image, channel='d'):
        if channel in [None, 'all']:
            return image

        if not isinstance(channel, Channel):
            channel = self.get_channel_from_name(channel)

        return channel.crop_image(image)

    def get_channel_from_name(self, channel_name):
        """Get the channel index belonging to a specific channel (name)
        If

        Parameters
        ----------
        channel : str or int
            The name or number of a channel

        Returns
        -------
        i: int
            The index of the channel to which the channel name belongs

        """
        for channel in self.channels:
            if channel_name in channel.names or channel_name == channel:
                return channel
        else:
            raise ValueError('Channel name not found')

    def get_channels_from_names(self, channel_names):
        """Get the channel index belonging to a specific channel (name)
        If

        Parameters
        ----------
        channel : str or int
            The name or number of a channel

        Returns
        -------
        i: int
            The index of the channel to which the channel name belongs

        """
        if channel_names in [None, 'all']:
            return self.channels

        if not isinstance(channel_names, list):
            channel_names = [channel_names]

        return [self.get_channel_from_name(channel_name) for channel_name in channel_names]

    def get_channel_indices_from_names(self, channel_names):
        channels = self.get_channels_from_names(channel_names)
        return [channel.index for channel in channels]

    def saveas_tif(self):
        tif_filepath = self.writepath.joinpath(self.name + '.tif')
        tif_filepath.unlink(missing_ok=True)

        for i in range(self.number_of_frames):
            frame = self.read_frames([i], apply_corrections=False, xarray=False)
            tifffile.imwrite(tif_filepath, frame, append=True)

            #     tifffile.imwrite(self.writepath.joinPath(f'{self.name}_fr{frame_number}.tif'), image,  photometric='minisblack')

    def make_projection_image(self, projection_type='average', frame_range=(0,20), apply_corrections=True, illumination=None, write=False,
                              return_image=True, flatten_channels=True, intensity_range=None, color_map='gray'):
        """ Construct a projection image
        Determine a projection image for a number_of_frames starting at start_frame.
        i.e. [start_frame, start_frame + number_of_frames)

        Parameters
        ----------
        projection_type : str
            'average' for average image
            'maximum' for maximum projection image
        start_frame : int
            Frame to start with
        number_of_frames : int
            Number of frames to average over
        write : bool
            If true, a tif file will be saved in writepath

        Returns
        -------
        np.ndarray
            2d image array with the projected image
        """

        frame_range = list(frame_range)
        # Make suitable for negative values
        if frame_range[0] > self.number_of_frames:
            raise ValueError(f'Invalid frame range {frame_range}')
        if frame_range[1] > self.number_of_frames:
            frame_range[1] = self.number_of_frames
            warnings.warn(f'Frame range exceeds available frames, used frame range {frame_range} instead')

        frame_indices = self.frame_indices.values[slice(*frame_range)]

        illumination_indices = self.get_illumination_indices_from_names(illumination)
        illumination_index = np.intersect1d(illumination_indices, self.illumination_indices_in_movie)[0]

        # Select frame_indices with illumination
        frame_indices = frame_indices[self.illumination_index_per_frame.values[frame_indices] == illumination_index]

        # Calculate sum of frames and find mean
        image = self.separate_channels(np.zeros((self.height, self.width)).astype('float32'))

        frame_indices_subsets = np.array_split(frame_indices, len(frame_indices) // self.chunk_size + 1)

        if projection_type == 'average':
            number_of_frames = len(frame_indices)
            # if len(frame_indices) > 100:
                # print(f'\n Making average image of {self.name}')
            with self:
                for frame_indices_subset in tqdm.tqdm(frame_indices_subsets, desc='Average image'):
                    # if len(frame_indices) > 100 and i % 13 == 0:
                    #     sys.stdout.write(
                    #         f'\r   Processing frame {frame_index} in {frame_indices[0]}-{frame_indices[-1]}')
                    frames = self.read_frames(frame_indices_subset, apply_corrections=apply_corrections,
                                              xarray=False, flatten_channels=False)
                    image = image + frames.sum(axis=0)
                #TODO: Check whether this is a good way to average, i.e. do the values not get too big.
            image = (image / number_of_frames).astype('float32')
        elif projection_type == 'maximum':
            # print(f'\n Making maximum projection image of {self.name}')
            with self:
                for frame_indices_subset in tqdm.tqdm(frame_indices_subsets, desc='Maximum projection image'):
                    # if i % 13 == 0:
                    #     sys.stdout.write(
                    #         f'\r   Processing frame {frame_index} in {frame_indices[0]}-{frame_indices[-1]}')
                    frames = self.read_frames(frame_indices_subset, xarray=False, flatten_channels=False)
                    image = np.maximum(image, frames.max(axis=0))
            # sys.stdout.write(f'\r   Processed frames {frame_indices[0]}-{frame_indices[-1]}\n')

        if write:
            filename = Movie.image_info_to_filename(self.name, fov_index=self.fov_index, projection_type=projection_type,
                                                    frame_range=frame_range, illumination=illumination_index, apply_corrections=apply_corrections)
            filepath = self.writepath.joinpath(filename)
            write_image = self.flatten_channels(image)
            if write in [True, 'tif']:
                if hasattr(self, 'pixel_size'):
                    resolution = 1/self.pixel_size
                else:
                    resolution = None
                tifffile.imwrite(filepath.with_suffix('.tif'), write_image,
                                 resolution=resolution,
                                 imagej=True,
                                 metadata={'unit': 'um',
                                           'axes': 'YX'}
                                 )
            elif write in ['png']:
                filepath = filepath.with_name(filepath.name + f'_v{intensity_range[0]}-{intensity_range[1]}')
                if intensity_range is None:
                    intensity_range = self.intensity_range
            # plt.imsave(filepath.with_suffix('.tif'), image, format='tif', cmap=colour_map, vmin=self.intensity_range[0], vmax=self.intensity_range[1])
                plt.imsave(filepath.with_suffix('.png'), write_image, vmin=intensity_range[0], vmax=intensity_range[1], cmap=color_map)

        if return_image:
            if flatten_channels:
                return self.flatten_channels(image)
            else:
                return image

    def make_projection_images(self, projection_type='average', frame_range=(0, 20)):
        # Perhaps put this in make_projection_image as a special type of cmap
        for illumination_index in range(self.number_of_illuminations_in_movie):
            image = self.make_projection_image(projection_type, frame_range=(0,20), illumination=illumination_index,
                                               write=True, return_image=True, flatten_channels=False)
            channel_images = []
            for channel_index in range(self.number_of_channels):
                channel_image = image[channel_index]
                channel_image = (channel_image - self.intensity_range[0]) / (self.intensity_range[1] - self.intensity_range[0]) # TODO: make separate intensity range for each channel
                channel_images.append(self.channels[channel_index].colour_map(channel_image, bytes=True))

            images_combined = np.hstack(channel_images)
            filename = Movie.image_info_to_filename(self.name, fov_index=self.fov_index, projection_type=projection_type,
                                                    frame_range=frame_range, illumination=illumination_index)
            filepath = self.writepath.joinpath(filename)
            plt.imsave(filepath.with_suffix('.png'), images_combined)


    def make_average_image(self, **kwargs):
        """ Construct an average image
        Determine average image for a number_of_frames starting at start_frame.
        i.e. [start_frame, start_frame + number_of_frames)

        Parameters
        ----------
        start_frame : int
            Frame to start with
        number_of_frames : int
            Number of frames to average over
        write : bool
            If true, the a tif file will be saved in the writepath

        Returns
        -------
        np.ndarray
            2d image array with the average image

        """
        return self.make_projection_image('average', **kwargs)

    def make_maximum_projection(self, **kwargs):
        """ Construct a maximum projection image
        Determine maximum projection image for a number_of_frames starting at start_frame.
        i.e. [start_frame, start_frame + number_of_frames)

        Parameters
        ----------
        start_frame : int
            Frame to start with
        number_of_frames : int
            Number of frames to average over
        write : bool
            If true, the a tif file will be saved in the writepath

        Returns
        -------
        np.ndarray
            2d image array with the maximum projection image
        """

        return self.make_projection_image('maximum', **kwargs)

    def show(self):
        return MoviePlotter(self)

    # Do we really need this?
    def determine_general_background_correction(self, method='median', frame_range=(0, 20), use_existing=False):
        if use_existing and 'general_background_correction' in self.corrections:
            return
        # self.temporal_background_correction = self.spatial_background_correction = None
        self.save_corrections(general_background_correction=None)

        frame_indices = self.frame_indices[slice(*frame_range)].values
        with self:
            frames = self.read_frames(frame_indices=frame_indices, apply_corrections=True, xarray=False)

        general_background_correction = xr.DataArray(0, dims=('illumination', 'channel'),
                                                      coords={'channel': self.channel_indices,
                                                              'illumination': self.illumination_indices},
                                                      name='general_background_correction')

        # corrections = self.corrections

        for illumination, channel in itertools.product(self.illumination_indices_in_movie,
                                                       np.array(self.channel_indices)):
            frame_indices_subset = (self.illumination_index_per_frame[frame_indices] == illumination).frame
            average_image = frames[frame_indices_subset, channel].mean(axis=0)

            # if 'flatfield_correction' in corrections:
            #     flatfield = corrections.flatfield_correction.sel(illumination=illumination, channel=channel).values
            # else:
            #     flatfield = None
            #
            # if 'darkfield_correction' in corrections:
            #     darkfield = corrections.darkfield_correction.sel(illumination=illumination, channel=channel).values
            # else:
            #     darkfield = None

            correction = determine_single_value_background_correction(average_image, method)#, flatfield, darkfield)

            # if 'temporal_illumination_correction' in corrections:
            #     correction /= corrections.temporal_illumination_correction[illumination, channel].mean().item()
            #
            # if 'temporal_background_correction' in corrections:
            #     correction -= corrections.temporal_background_correction[frame_indices_subset, channel].mean().item()
            #
            # if 'spatial_background_correction' in corrections:
            #     correction -= corrections.spatial_background_correction[illumination, channel].mean().item()

            general_background_correction[dict(illumination=illumination, channel=channel)] = correction

        # self.general_background_correction = general_background_correction
        # self.save_corrections(general_background_correction=general_background_correction,
        #                       temporal_background_correction=None, spatial_background_correction=None)
        self.save_corrections(general_background_correction=general_background_correction)

    def determine_temporal_background_correction(self, method='median', use_existing=False):
        if use_existing and 'temporal_background_correction' in self.corrections:
            return

        # self.spatial_background_correction = None
        self.save_corrections(temporal_illumination_correction=None,
                              temporal_background_correction=None,
                              spatial_background_correction=None,
                              general_background_correction=None)

        frames = self.read_frames(frame_indices=None, apply_corrections=True, xarray=False)

        temporal_background_correction = xr.DataArray(0, dims=('frame', 'channel'),
                                                      coords={'frame': self.frame_indices,
                                                              'channel': self.channel_indices},
                                                      name='temporal_background_correction')

        # corrections = self.corrections

        for illumination, channel in itertools.product(self.illumination_indices_in_movie, np.array(self.channel_indices)):

            frame_indices_subset = (self.illumination_index_per_frame==illumination).frame
            frames_subset = frames[frame_indices_subset, channel]

            # if 'flatfield_correction' in corrections:
            #     flatfield = corrections.flatfield_correction.sel(illumination=illumination, channel=channel).values
            # else:
            #     flatfield = None
            #
            # if 'darkfield_correction' in corrections:
            #     darkfield = corrections.darkfield_correction.sel(illumination=illumination, channel=channel).values
            # else:
            #     darkfield = None

            correction = determine_temporal_background_correction(frames_subset, method)#, flatfield, darkfield)

            # if 'general_background_correction' in corrections:
            #     correction -= corrections.general_background_correction[illumination, channel].item()

            # if 'spatial_background_correction' in corrections:
            #     correction -= corrections.spatial_background_correction[illumination, channel].mean().item()

            temporal_background_correction[dict(frame=frame_indices_subset, channel=channel)] = correction

        # self.temporal_background_correction = temporal_background_correction
        # self.save_corrections(temporal_background_correction=temporal_background_correction,
        #                       spatial_background_correction=None)
        self.save_corrections(temporal_background_correction=temporal_background_correction)#,
                              # spatial_background_correction=None,
                              # general_background_correction=None)

    def determine_spatial_background_correction(self, method='median_filter', frame_range=(0, 20), use_existing=False,
                                                **kwargs):
        if use_existing and 'spatial_background_correction' in self.corrections:
            return

        self.save_corrections(spatial_background_correction=None, general_background_correction=None)

        frame_indices = self.frame_indices[slice(*frame_range)].values
        with self:
            frames = self.read_frames(frame_indices=frame_indices, apply_corrections=True, xarray=False)

        spatial_background_correction = xr.DataArray(np.zeros((self.number_of_illuminations,) + frames.shape[1:]),
                                                     dims=('illumination', 'channel', 'y', 'x'),
                                                     coords={'illumination': self.illumination_indices,
                                                             'channel': self.channel_indices, },
                                                     name='spatial_background_correction')

        # corrections = self.corrections

        for illumination, channel in itertools.product(self.illumination_indices_in_movie,
                                                       np.array(self.channel_indices)):
            frame_selection = (self.illumination_index_per_frame[frame_indices] == illumination).values
            average_image = frames[frame_selection, channel].mean(axis=0)
            # if 'flatfield_correction' in corrections:
            #     flatfield = corrections.flatfield_correction.sel(illumination=illumination, channel=channel).values
            # else:
            #     flatfield = None
            #
            # if 'darkfield_correction' in corrections:
            #     darkfield = corrections.darkfield_correction.sel(illumination=illumination, channel=channel).values
            # else:
            #     darkfield = None

            correction = determine_spatial_background_correction(average_image, method, **kwargs) # flatfield, darkfield, **kwargs)

            # if 'general_background_correction' in corrections:
            #     correction -= corrections.general_background_correction[illumination, channel].item()

            # if 'temporal_background_correction' in corrections:
            #     correction -= corrections.temporal_background_correction[frame_indices_subset, channel].mean().item()

            spatial_background_correction[dict(illumination=illumination, channel=channel)] = correction

        # self.spatial_background_correction = spatial_background_correction
        # self.save_corrections(spatial_background_correction=spatial_background_correction)
        self.save_corrections(spatial_background_correction=spatial_background_correction)#,
                              # temporal_background_correction=None,
                              #general_background_correction=None)
    @property
    def corrections(self):
        if hasattr(self, 'fov_index') and self.fov_index is not None:
            corrections_filepath = self.filepath.with_name(self.name + f'_fov{self.fov_index:03d}' + '_corrections.nc')
        else:
            corrections_filepath = self.filepath.with_name(self.name + '_corrections.nc')

        if corrections_filepath.exists():
            corrections = xr.load_dataset(corrections_filepath, engine='h5netcdf')
        else:
            corrections = xr.Dataset()
        corrections = corrections.merge(self._common_corrections, compat='override')
        return corrections

    # def load_corrections(self):
    #     corrections_filepath = self.filepath.with_name(self.name + '_corrections.nc')
    #     if corrections_filepath.exists():
    #         corrections = xr.load_dataset(corrections_filepath, engine='h5netcdf')
    #         # for key, correction in corrections.data_vars.items():
    #         #     self.__setattr__(key, correction)
    #     return corrections

    @property
    def corrections_filepath(self):
        if hasattr(self, 'fov_index') and self.fov_index is not None:
            corrections_filepath = self.filepath.with_name(self.name + f'_fov{self.fov_index:03d}' + '_corrections.nc')
        else:
            corrections_filepath = self.filepath.with_name(self.name + '_corrections.nc')
        return corrections_filepath

    def reset_corrections(self):
        self.corrections_filepath.unlink(missing_ok=True)

    def save_corrections(self, **kwargs):
        corrections_filepath = self.corrections_filepath
        if corrections_filepath.exists():
            corrections = xr.load_dataset(corrections_filepath, engine='h5netcdf')
        else:
            corrections = xr.Dataset()
        for name, correction in kwargs.items():
            # correction = getattr(self, name)
            if correction is None:
                corrections = corrections.drop_vars(name, errors='ignore')
            else:
                corrections[name] = correction
        corrections.to_netcdf(corrections_filepath, mode='w', engine='h5netcdf')

#     def apply_corrections(self, frames, frame_indices):
#
#
#         return apply_corrections(frames, frame_indices, self.darkfield_correction.values,
#                                  self.flatfield_correction.values, self.illumination_correction.values,
#                                  self.background_correction.values)
#
# # @njit
    def apply_corrections(self, frames, frame_indices):
        illumination_indices = self.illumination_index_per_frame[frame_indices]
        frames = frames.astype(np.float32)
        corrections = self.corrections
        for illumination_index in np.unique(illumination_indices):
            frame_indices_with_illumination = np.array(illumination_indices == illumination_index)

            if 'darkfield_correction' in corrections:
                frames[frame_indices_with_illumination] -= corrections.darkfield_correction.values[None, illumination_index]

            if 'flatfield_correction' in corrections:
                frames[frame_indices_with_illumination] /= corrections.flatfield_correction.values[None, illumination_index]

            if 'temporal_illumination_correction' in corrections:
                frames[frame_indices_with_illumination] /= \
                    corrections.temporal_illumination_correction.values[frame_indices][frame_indices_with_illumination, None, None, None]

            if 'temporal_background_correction' in corrections:
                frames[frame_indices_with_illumination] -= \
                    corrections.temporal_background_correction.values[frame_indices][frame_indices_with_illumination, :, None, None]

            if 'spatial_background_correction' in corrections:
                frames[frame_indices_with_illumination] -= corrections.spatial_background_correction.values[None,
                                                           illumination_index, :, :, :]

            if 'general_background_correction' in corrections:
                frames[frame_indices_with_illumination] -= corrections.general_background_correction.values[None, illumination_index, :, None, None]

        return frames

    def show_correction(self, correction_name, save=True, **kwargs):
        correction = self.corrections[correction_name]
        number_of_illuminations = len(correction.illumination)
        figure, axes = plt.subplots(1, number_of_illuminations+1, gridspec_kw=dict(width_ratios=(4,)*number_of_illuminations + (0.15,)),
                                    figsize=(4*number_of_illuminations+0.15, 4), layout='tight')
        for i, illumination_index in enumerate(correction.illumination):
            axes[i].axis('off')
            image = axes[i].imshow(self.flatten_channels(correction.sel(illumination=illumination_index)), **kwargs)
            axes[i].set_title(f'Illumination {illumination_index.item()}', fontsize=8)

        from mpl_toolkits.axes_grid1 import make_axes_locatable
        # divider = make_axes_locatable(axes[-1])
        # cax = divider.append_axes("right", "4%", pad="15%")
        cax = axes[-1]
        figure.colorbar(image, aspect=50, cax=cax)
        cax.set_ylabel('Intensity (a.u.)')
        if correction_name == 'flatfield_correction':
            cax.set_ylabel('Correction factor')

        # cax.axes.ticklabel_format(scilimits=(0, 0))
        for spine in cax.spines.values():
            spine.set_visible(False)

        figure.suptitle(f'{self.name} - {correction_name}', fontsize=8)

        if save:
            figure.savefig(self.filepath.with_name(f'{self.name} - {correction_name}.png'), bbox_inches='tight')


class Channel:
    def __init__(self, movie, name, short_name, other_names=[], colour_map=None):
        self.movie = movie
        self.name = name
        self.short_name = short_name
        self.other_names = other_names
        if colour_map is None:
            channel_colour = \
            list({'green', 'red', 'blue'}.intersection([self.name, self.short_name] + self.other_names))[0]
            self.colour_map = make_colour_map(channel_colour)

    def __repr__(self):
        return (f'{self.__class__.__name__}({self.name})')

    @property
    def names(self):
        return [self.index, str(self.index), self.name, self.short_name] + self.other_names

    @property
    def index(self):
        try:
            return self.movie.channels.index(self)
        except:
            pass

    @property
    def location(self):
        return [int(i) for i in np.where(self.movie.channel_arrangement == self.index)]

    @property
    def width(self):
        return self.movie.width // self.movie.channel_arrangement.shape[2]

    @property
    def height(self):
        return self.movie.height // self.movie.channel_arrangement.shape[1]
        # for frame_index, frame in enumerate(self.channel_arrangement):
        #     for y_index, y in enumerate(frame):
        #         try:
        #             x_index = y.index(channel_index)
        #             return frame_index, y_index, x_index
        #         except ValueError:
        #             pass

    @property
    def dimensions(self):
        return np.array([self.width, self.height])

    @property
    def origin(self):
        return [self.width * self.location[2],
                self.height * self.location[1]]

    @property
    def boundaries(self):
        # channel_boundaries: np.array
        # #         Formatted as two coordinates, with the lowest and highest x and y values respectively
        horizontal_boundaries = np.array([0, self.width]) + self.width * self.location[2]
        vertical_boundaries = np.array([0, self.height]) + self.height * self.location[1]
        return np.vstack([horizontal_boundaries, vertical_boundaries]).T

    @property
    def vertices(self):
        #     channel_vertices : np.array
        #         Four coordinates giving the four corners of the channel
        #         Coordinates form a closed shape
        channel_vertices = np.array([self.origin, ] * 4)
        channel_vertices[[1, 2], 0] += self.width
        channel_vertices[[2, 3], 1] += self.height
        return channel_vertices

    def crop_image(self, image):
        return image[self.boundaries[0, 1]:self.boundaries[1, 1],
               self.boundaries[0, 0]:self.boundaries[1, 0]]

    def crop_images(self, images):
        return images[:, self.boundaries[0, 1]:self.boundaries[1, 1],
               self.boundaries[0, 0]:self.boundaries[1, 0]]


class MoviePlotter:
    # Adapted from Matplotlib Image Slices Viewer
    def __init__(self, movie):
        fig, ax = plt.subplots(1, 1)
        fig.canvas.mpl_connect('scroll_event', self.on_scroll)
        plt.show()

        self.ax = ax
        ax.set_title('use scroll wheel to navigate images')

        self.movie = movie
        self.slices, rows, cols = (movie.number_of_frames, movie.height, movie.width)
        self.ind = self.slices // 2

        self.im = ax.imshow(self.movie.read_frame(self.ind, flatten_channels=True, xarray=False))
        self.update()

    def on_scroll(self, event):
        print("%s %s" % (event.button, event.step))
        if event.button == 'up':
            self.ind = (self.ind + 1) % self.slices
        else:
            self.ind = (self.ind - 1) % self.slices
        self.update()

    def update(self):
        self.im.set_data(self.movie.read_frame(self.ind, flatten_channels=True, xarray=False))
        self.ax.set_ylabel('slice %s' % self.ind)
        self.im.axes.figure.canvas.draw()


def make_colour_map(colour, N=256):
    values = np.zeros((N, 3))
    if colour == 'grey':
        values[:, 0] = values[:, 1] = values[:, 2] = np.linspace(0, 1, N)
    elif colour == 'red':
        values[:, 0] = np.linspace(0, 1, N)
    elif colour == 'green':
        values[:, 1] = np.linspace(0, 1, N)
    elif colour == 'blue':
        values[:, 2] = np.linspace(0, 1, N)
    else:
        values[:, 0] = values[:, 1] = values[:, 2] = np.linspace(0, 1, N)

    return ListedColormap(values)


def expand_axes(frames, expand_into, from_axes=-1, to_axes=None, new_axes_positions=[], inverse=False, squeeze=False):
    if isinstance(expand_into, int):
        expand_into = (expand_into,)
    if isinstance(from_axes, int):
        from_axes = (from_axes,)
    if isinstance(to_axes, int) or to_axes is None:
        to_axes = (to_axes,) * len(from_axes)

    from_axes = list(from_axes)
    to_axes = list(to_axes)

    if inverse:
        expand_into = expand_into[::-1]
        from_axes, to_axes = to_axes[::-1], from_axes[::-1]

    ndim = frames.ndim

    # new_axes_created = 0
    for i, (from_axis, to_axis) in enumerate(zip(from_axes, to_axes)):
        # if from_axis is None:
        #     from_axes[i] = ndim-1
        if from_axis < 0:
            from_axes[i] = range(ndim)[from_axis]
        #
        # if to_axis is None:
        #     # if combine_new_axes and new_axes_created > 0:
        #     #     to_axes[i] = ndim
        #     # else:
        #     new_axes_positions.append(ndim)
        #     new_axes_created += 1
        if -ndim <= to_axis < 0:
            to_axes[i] = range(ndim)[to_axis]

        elif to_axis < -ndim:  # or to_axis > ndim-1:
            if to_axis not in new_axes_positions:
                new_axes_positions.append(to_axis)
            if to_axis < 0:
                to_axes[i] = ndim + new_axes_positions.index(to_axis)

            # to_axes[i] = None
            # new_axes_created += 1

    for i, (n, from_axis, to_axis) in enumerate(zip(expand_into, from_axes, to_axes)):
        # if frames.shape[-1] % n > 0:
        #     raise ValueError('Cannot split into equal parts')
        if to_axis > frames.ndim - 1:
            frames = np.moveaxis(frames, from_axis, -1)
            frames = frames.reshape(*frames.shape[:-1], n, frames.shape[-1] // n)
            frames = np.moveaxis(frames, -1, from_axis)
        elif inverse:
            frames = np.moveaxis(frames, [from_axis, to_axis], [-2, -1])
            frames = frames.reshape(*frames.shape[:-2], frames.shape[-2] // n, frames.shape[-1] * n)
            frames = np.moveaxis(frames, [-2, -1], [from_axis, to_axis])
        else:
            frames = np.moveaxis(frames, [from_axis, to_axis], [-1, -2])
            frames = frames.reshape(*frames.shape[:-2], frames.shape[-2] * n, frames.shape[-1] // n)
            frames = np.moveaxis(frames, [-1, -2], [from_axis, to_axis])

    if inverse and squeeze:
        for from_axis in np.sort(np.unique(from_axes))[::-1]:
            if frames.shape[from_axis] <= 1:
                frames = frames.squeeze(axis=from_axis)
        # frames = np.moveaxis(frames, from_axis, -1)

    if new_axes_positions:
        frames = np.moveaxis(frames, -np.arange(len(new_axes_positions))[::-1] - 1, new_axes_positions)

    # Test code
    # start = time.time()
    # a = np.stack(np.split(frames, 2, axis=2), axis=-1)
    # b = np.concatenate(np.split(a, 2, axis=1), axis=-1)
    # print(time.time() - start)
    #
    # start = time.time()
    # bb = split_image_channels(frames, (2, 2), axes=(1, 2), combine_new_axes=False)
    # print(time.time() - start)

    return frames


import numpy as np
import matplotlib.pyplot as plt
from papylio.movie.movie import Movie


class BinaryMovie(Movie):
    extensions = ['.bin']

    def __init__(self, arg, *args, **kwargs):
        super().__init__(arg, *args, **kwargs)
        
        self.writepath = self.filepath.parent
        self.name = self.filepath.with_suffix('').name
        
        #determine 8 bits or 16 bits
        # self.bitdepth = 16 if (self.filepath.name[-7:-4]=='_16') else 8

        self.threshold = {  'view':             (0,200),
                            'point-selection':  (45,25)
                            }


        self.width = 250
        self.height = 250
        self.number_of_frames = 200
        self.illumination_arrangement = np.array([1, 0])
        self.channel_arrangement = np.array([[[1]], [[0]]])

        # self.read_header()

        self.create_frame_info() # Possibly move to Movie later on

    def _read_header(self):
        pass

    def _read_frame(self, frame_number):
        # t = time.time()
        # if (frame_number < 0) or (frame_number >= self.number_of_frames):
        #     raise ValueError('Frame number out of range')
        #
        # start_byte = frame_number*self.bytes_per_frame
        #
        # with self.filepath.open('rb') as bin:
        #     bin.seek(start_byte)
        #     data = np.fromfile(bin, dtype=self.dtype, count=self.pixels_per_frame)
        #     image = data.reshape((self.width, self.height))
        #
        # return image

        return self.read_frames(frame_number, 1).squeeze(0)

    def read_frames(self, start_frame=0, number_of_frames=None):
        if number_of_frames is None:
            number_of_frames = self.number_of_frames

        for frame_number in (start_frame, start_frame+number_of_frames-1):
            if (frame_number < 0) or (frame_number >= self.number_of_frames):
                raise ValueError('Frame number out of range')

        start_byte = start_frame * self.bytes_per_frame

        with self.filepath.open('rb') as bin_file:
            bin_file.seek(start_byte)
            data = np.fromfile(bin_file, dtype=self.data_type, count=self.pixels_per_frame*number_of_frames)
            image = data.reshape((number_of_frames, self.width, self.height))

        return image

if __name__ == "__main__":
    movie = BinaryMovie(r'.\Example_data\binary\movie.bin')
    test = movie.read_frames(2,10)
    plt.imshow(test[0])
    movie.make_projection_images()





import numpy as np
import xarray as xr
from papylio.plugins.sequencing.plotting import double_mutations, plot_double_mutations
import pytest

@pytest.fixture
def general_sequence():
    return 'ACTGACTG'

@pytest.fixture
def simulated_data(general_sequence):
    dm = double_mutations(general_sequence, add_reference=True)
    da = xr.DataArray(np.arange(len(dm))[::-1], dims=('sequence',), coords={'sequence': dm})
    da.name = 'Test dataset'
    return da

def test_plot_double_mutations(general_sequence, simulated_data):
    # Without annotations
    plot_double_mutations(general_sequence, simulated_data)

    # With annotations, i.e. reporting the number
    plot_double_mutations(general_sequence, simulated_data, da_annotation=simulated_data)


import pytest
from PySide2 import QtWidgets
import sys
from papylio.trace_correction import TraceCorrectionWindow


@pytest.fixture
def experiment(shared_datadir):
    from papylio import Experiment
    return Experiment(shared_datadir / 'BN_TIRF_output_test_file')


@pytest.fixture
def file(experiment):
    return experiment.files[0]


def test_trace_plot(file):
    # app = QtWidgets.QApplication.instance()
    # if not app:
    #     app = QtWidgets.QApplication(sys.argv)
    frame = TraceCorrectionWindow(file.intensity)
    # frame.show()
    # frame.activateWindow()
    # frame.raise_()
    # app.exec_()



import numpy as np
import matplotlib.pyplot as plt
from objectlist import ObjectList
from scipy import stats
from tqdm import tqdm
from matplotlib.ticker import MaxNLocator

from objectlist import ObjectList
from matchpoint.core import distance_threshold_from_number_of_matches


# TODO: Make sure that the collection can only contain MatchPoint objects
class MappingCollection(ObjectList):
    dimension_number = {0: 0, 1: 1, 'x': 0, 'y': 1}
    dimension_string = {0: 'x', 1: 'y', 'x': 'x', 'y': 'y'}

    @property
    def translations(self):
        return np.vstack(self.translation)

    @property
    def unit_string(self):

        if self[0].destination_unit is not None and len(np.unique(self.destination_unit)) == 1:
            destination_unit_string = ' (' + self[0].destination_unit + ')'
        else:
            destination_unit_string = ''

        unit_string = {'translation': destination_unit_string, 'rotation': ' (rad)', 'scale': '',
                       'shear': ' (rad)'}

        return unit_string

    def plot_parameter(self, parameter, dimension=0, label_name='Label', indices=None, figure=None, save=False, **kwargs):
        data = np.vstack(getattr(self, parameter))[:, self.dimension_number[dimension]]

        all_indices = range(len(self))
        if indices is None:
            indices = all_indices

        if figure is None:
            figure, axis = plt.subplots()
        else:
            axis = figure.gca()

        axis.scatter(indices, data[indices], **kwargs)
        axis.set_xlabel(label_name)
        axis.set_ylabel(parameter.capitalize() + ' ' + self.dimension_string[dimension] + self.unit_string[parameter])
        axis.set_xticks(all_indices)
        axis.set_xticklabels(self.label, rotation=45, ha="right")

        figure.tight_layout()
        if save:
            figure.savefig(self[0].save_path.joinpath(f'{parameter.capitalize()}-{self.dimension_string[dimension]}.png'))

        return figure, axis

    def scatter_parameters(self, parameter_x, parameter_y, dimension_x=0, dimension_y=0, indices=None, figure=None, save=False, **kwargs):
        data_x = np.vstack(getattr(self, parameter_x))[:, self.dimension_number[dimension_x]]
        data_y = np.vstack(getattr(self, parameter_y))[:, self.dimension_number[dimension_y]]

        if indices is None:
            indices = np.arange(len(data_x))

        if figure is None:
            figure, axis = plt.subplots()
        else:
            axis = figure.gca()

        axis.scatter(data_x[indices], data_y[indices], **kwargs)
        axis.set_xlabel(parameter_x.capitalize() + ' ' + self.dimension_string[dimension_x] + self.unit_string[parameter_x])
        axis.set_ylabel(parameter_y.capitalize() + ' ' + self.dimension_string[dimension_y] + self.unit_string[parameter_y])
        for label, x, y in zip(self.label[indices], data_x[indices], data_y[indices]):
            axis.annotate(label, (x, y))

        figure.tight_layout()
        if save:
            figure.savefig(self[0].save_path.joinpath(f'{parameter_y.capitalize()}-{self.dimension_string[dimension_y]}_vs_'
                                                      f'{parameter_x}-{self.dimension_string[dimension_x]}.png'))

        return figure, axis

    # def plot_translations(self, dimension=None, save=False, **kwargs):
    #     unit_string = self.destination_unit_string
    #
    #     fig, ax = plt.subplots()
    #     if dimension is None:
    #         ax.scatter(*self.translations.T)
    #         ax.set_xlabel('Translation x' + unit_string)
    #         ax.set_ylabel('Translation y' + unit_string)
    #         for label, translation in zip(self.label, self.translations):
    #             ax.annotate(label, translation)
    #     elif dimension in ['x', 0]:
    #         ax.scatter(range(len(self)), self.translations[:, 0])
    #         ax.set_xlabel('Label')
    #         ax.set_ylabel('Translation x' + unit_string)
    #         ax.set_xticks(range(len(self)))
    #         ax.set_xticklabels(self.label, rotation=45, ha="right")
    #     elif dimension in ['y', 1]:
    #         ax.scatter(range(len(self)), self.translations[:,1])
    #         ax.set_xlabel('Label')
    #         ax.set_ylabel('Translation y' + unit_string)
    #         ax.set_xticks(range(len(self)))
    #         ax.set_xticklabels(self.label, rotation=45, ha="right")
    #     else:
    #         raise ValueError("Dimension can be 'x', 'y' or None.")
    #
    #     fig.tight_layout()
    #     if save:
    #         fig.savefig(self[0].save_path.joinpath(f'Tile_translations.png'), **kwargs)
    #
    #     return fig, ax

    def fit_translations(self, indices, save=False):
        all_indices = np.arange(len(self))
        if indices is None:
            indices = all_indices
        res = stats.linregress(*self.translations[indices].T)
        figure, axis = self.scatter_parameters('translation', 'translation', 'x', 'y')
        self.scatter_parameters('translation', 'translation', 'x', 'y', indices=indices, figure=figure, c='orange')
        xlim, ylim = axis.get_xlim(), axis.get_ylim()
        x = np.array(xlim)
        axis.plot(x, res.slope * x + res.intercept, c='orange')
        axis.set_xlim(xlim), axis.set_ylim(ylim)
        figure.tight_layout()
        if save:
            figure.savefig(self[0].save_path.joinpath(f'Translations_fit.png'))

        # Plot and fit x translation vs tile number
        res_x = stats.linregress(indices, self.translations[indices, 0])
        figure, axis = self.plot_parameter('translation', dimension='x')
        self.plot_parameter('translation', dimension='x', indices=indices, figure=figure, c='orange')
        axis.plot(all_indices, res_x.slope * all_indices + res_x.intercept)
        if save:
            figure.savefig(self[0].save_path.joinpath(f'Translation-x_vs_index_fit.png'))

        # Plot and fit y translation vs tile number
        res_y = stats.linregress(indices, self.translations[indices, 1])
        figure, axis = self.plot_parameter('translation', dimension='y')
        self.plot_parameter('translation', dimension='y', indices=indices, figure=figure, c='orange')
        axis.plot(all_indices, res_y.slope * all_indices + res_y.intercept)
        if save:
            figure.savefig(self[0].save_path.joinpath(f'Translation-y_vs_index_fit.png'))

        return res_x, res_y

    def estimate_translations(self, **kwargs):
        res_x, res_y = self.fit_translations(**kwargs)

        for i, mapping in enumerate(self):
            mapping.initial_transformation = type(mapping.transformation)(matrix=mapping.transformation.params) # To make a copy
            new_translation = [res_x.slope * i + res_x.intercept, res_y.slope * i + res_y.intercept]
            mapping.transformation.params[0:2, 2] = new_translation
            mapping.calculate_inverse_transformation()

    # TODO: Extract parameters / statistics
    # columns = pd.MultiIndex.from_product([['Single-molecule', 'Sequencing'], ['Total', 'Matched', 'Fraction']])
    # df = pd.DataFrame(columns=columns)
    # for tile, mapping in tile_mappings_calculated.items():
    #     mapping.source_distance_threshold = 5  # um
    #     pm = mapping.number_of_matched_points
    #     fs, fd = mapping.fraction_of_points_matched('source')
    #     df.loc[tile, :] = [len(mapping.source_cropped), pm, fs, len(mapping.destination), pm, fd]
    # df.index.name = 'Tile'
    # df.to_excel(analysis_path.joinpath(f'Fraction_of_points_matched_source_threshold_5um.xlsx'))

    def find_distance_threshold(self, method='single_match_optimization', **kwargs):
        if method == 'single_match_optimization':
            self.single_match_optimization(**kwargs)
        else:
            raise ValueError('Unknown method')

    def single_matches_over_radius(self, maximum_radius=20, number_of_steps=100):
        radii = np.linspace(0, maximum_radius, number_of_steps)
        number_of_pairs = self.number_of_single_matches_for_radii(radii)
        number_of_pairs_summed = np.vstack(number_of_pairs).sum(axis=0)
        return radii, number_of_pairs_summed

    def single_match_optimization(self, maximum_radius=20, number_of_steps=100, plot=True):
        radii, number_of_pairs = self.single_matches_over_radius(maximum_radius, number_of_steps)
        self.destination_distance_threshold = distance_threshold_from_number_of_matches(radii, number_of_pairs, plot=plot)



# -*- coding: utf-8 -*-
"""
Created on Thu Apr  9 00:41:20 2020

@author: iason
"""

if __name__ == '__main__':
    import sys
    from pathlib import Path
    p = Path(__file__).parents[2]  # include two levels up directory where the papylio package lies
    sys.path.insert(0, str(p))

import numpy as np
import pandas as pd
from papylio.experiment import Experiment

import matplotlib.pyplot as plt
import seaborn as sns

def plot_fret_histogram(molecules, Emin=0.05, Emax=0.8, Irmin=40, Iroff=0,
                        Igoff=0, **hist_kwargs):
    plt.style.use('seaborn-dark')
    plt.style.use('seaborn-colorblind')
    data = np.concatenate([molecule.E(Irmin, Iroff, Igoff) for molecule in molecules])
    data = data[data>Emin]
    data = data[data<Emax]
    centers, bins, _ = plt.hist(data, **hist_kwargs)


def get_fret_for_offtimes(molecules, trace_type='red', Emin=0.05, Emax=1,
                        Irmin=40, Iroff=0, Igoff=0, fret_max=False):
    '''
    Function that gives the FRET trace with FRET calculcated only for the
    selected offtime interval. The FRET value outside this interval is put to
    zero
    '''
    fret_new = []
    for mol in list(molecules):
        if mol.steps is None:
            continue
        print(mol.index)
        time_axis = mol.file.time
        # select the times in which steps are taken for the particular trace_type
        times = mol.steps.time.values[mol.steps.trace == trace_type]
        indices = np.empty(times.size, dtype=int)
        for i, t in enumerate(times):
            idx = find_nearest(time_axis, t)
            indices[i] = idx
        fret = mol.E(Irmin, Iroff, Igoff)
        idx_start = indices[::2]
        idx_stop = indices[1::2]

        if fret_max:
            fret_reduced = []
            for i in range(idx_start.size):
                try:
                    maximum = np.max(fret[idx_start[i]+1:idx_stop[i]])
                except ValueError:
                    pass
                fret_new.append(maximum)

        else:
            mask = np.zeros(fret.size, dtype=int)
            for i in range(idx_start.size):
                mask[idx_start[i]+1:idx_stop[i]] = 1
            # Multiply by zero all the fret values outside the offtime intervals
            fret_reduced = np.multiply(fret, mask)

            fret_new.append(fret_reduced)

    fret_new = np.array(fret_new)
    fret_new = fret_new[fret_new > 0]

    return fret_new


def find_nearest(array, value):
    # array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return int(idx)


if __name__ == '__main__':

    main_path = r'C:/Users/iason/Desktop/traceanalysis/papylio/traces/test_data/DNA08'

    exp = Experiment(main_path)
    file = exp.files[0]
    filename = main_path + '/hel21_dwells_data.xlsx'
    dwell_data =  pd.read_excel(filename, index_col=[0, 1], dtype={'kon' :np.str})
    # fret_hist = plot_fret_histogram(file.molecules, bins=100)
    mol = file.molecules[0]
    fret = get_fret_for_offtimes(file.molecules)
    plt.hist(fret, bins=50)




import pytest
from papylio.plugins.sequencing.sequencing_data import parse_sam, fastq_generator, add_sequence_data_to_dataset, SequencingData


@pytest.fixture
def read1_fastq_filepath(shared_datadir):
    return shared_datadir / 'sequencing' / 'read1.fastq'


@pytest.fixture
def index1_fastq_filepath(shared_datadir):
    return shared_datadir / 'sequencing' / 'index1.fastq'


@pytest.fixture
def sam_filepath(shared_datadir):
    return shared_datadir / 'sequencing' / 'aligned.sam'


@pytest.fixture
def nc_filepath(sam_filepath):
    test_parse_sam(sam_filepath)
    return sam_filepath.with_suffix('.nc')


@pytest.fixture
def sequencing_data(nc_filepath):
    return SequencingData(file_path=nc_filepath)


def test_parse_sam(sam_filepath):
    extract_sequence_subset = [30, 31, 56, 57, 82, 83, 108, 109]
    parse_sam(sam_filepath, remove_duplicates=True, add_aligned_sequence=True, extract_sequence_subset=extract_sequence_subset,
              chunksize=10, write_csv=False, write_nc=True, write_filepath=None)


def test_fastq_generator(read1_fastq_filepath):
    fqg = fastq_generator(read1_fastq_filepath, 100)
    next(fqg)


def test_add_sequence_data_to_dataset(nc_filepath, index1_fastq_filepath):
    add_sequence_data_to_dataset(nc_filepath, index1_fastq_filepath, 'index1')


def test_sequencing_data_coordinates(sequencing_data):
    sequencing_data.coordinates


# def test_unlim_vs_lim_dim_netcdf4():
#     import netCDF4
#     import numpy as np
#
#     ds_unlim = netCDF4.Dataset(r'D:\Test\ds_unlim.nc', 'w')
#     ds_unlim.createDimension('test_dim', None)
#     ds_unlim.createVariable('test_var', np.uint32, ('test_dim',), chunksizes=(1,))
#     ds_unlim['test_var'][:] = np.arange(10 ** 6)
#     ds_unlim.close()
#
#     ds_lim = netCDF4.Dataset(r'D:\Test\ds_lim.nc', 'w')
#     ds_lim.createDimension('test_dim', 10**6)
#     ds_lim.createVariable('test_var', np.uint32, ('test_dim',))
#     ds_lim['test_var'][:] = np.arange(10 ** 6)
#     ds_lim.close()
#
#
#     import xarray as xr
#     import time
#     ds_unlim = xr.open_dataset(r'D:\Test\ds_unlim.nc')
#     start = time.time()
#     ds_unlim['test_var'].load()
#     print(time.time()-start)
#     ds_unlim.close()
#
#     ds_lim = xr.open_dataset(r'D:\Test\ds_lim.nc')
#     start = time.time()
#     ds_lim['test_var'].load()
#     print(time.time()-start)
#     ds_lim.close()